@@ -18,12 +18,12 @@
 # -- Project information -----------------------------------------------------
 
 project = 'pgcopydb'
-copyright = '2022, Dimitri Fontaine'
+copyright = '2022-2023, Dimitri Fontaine'
 author = 'Dimitri Fontaine'
 
 # The full version, including alpha/beta/rc tags
-version = '0.10'
-release = '0.10'
+version = '0.13'
+release = '0.13'
 
 
 # -- General configuration ---------------------------------------------------
@@ -60,6 +60,7 @@ html_theme = "sphinx_rtd_theme"
 # so a file named "default.css" will overwrite the builtin "default.css".
 html_static_path = ['_static']
 
+latex_engine = 'xelatex'
 
 #
 # https://stackoverflow.com/questions/9899283/how-do-you-change-the-code-example-font-size-in-latex-pdf-output-with-sphinx
@@ -112,6 +113,13 @@ man_pages = [
         [author],
         1,
     ),
+    (
+        "ref/pgcopydb_compare",
+        "pgcopydb compare",
+        "pgcopydb compare",
+        [author],
+        1,
+    ),
     (
         "ref/pgcopydb_copy",
         "pgcopydb copy",
