@@ -32,17 +32,24 @@
 	"  --split-tables-larger-than Same-table concurrency size threshold\n" \
 	"  --drop-if-exists           On the target database, clean-up from a previous run first\n" \
 	"  --roles                    Also copy roles found on source to target\n" \
+	"  --no-role-passwords        Do not dump passwords for roles\n" \
 	"  --no-owner                 Do not set ownership of objects to match the original database\n" \
 	"  --no-acl                   Prevent restoration of access privileges (grant/revoke commands).\n" \
 	"  --no-comments              Do not output commands to restore comments\n" \
 	"  --skip-large-objects       Skip copying large objects (blobs)\n" \
 	"  --skip-extensions          Skip restoring extensions\n" \
+	"  --skip-ext-comments        Skip restoring COMMENT ON EXTENSION\n" \
+	"  --skip-collations          Skip restoring collations\n" \
+	"  --skip-vacuum              Skip running VACUUM ANALYZE\n" \
+	"  --requirements <filename>  List extensions requirements\n" \
 	"  --filters <filename>       Use the filters defined in <filename>\n" \
+	"  --fail-fast                Abort early in case of error\n" \
 	"  --restart                  Allow restarting when temp files exist already\n" \
 	"  --resume                   Allow resuming operations after a failure\n" \
 	"  --not-consistent           Allow taking a new snapshot on the source database\n" \
 	"  --snapshot                 Use snapshot obtained with pg_export_snapshot\n" \
 	"  --follow                   Implement logical decoding to replay changes\n" \
+	"  --plugin                   Output plugin to use (test_decoding, wal2json)\n" \
 	"  --slot-name                Use this Postgres replication slot name\n" \
 	"  --create-slot              Create the replication slot\n" \
 	"  --origin                   Use this Postgres replication origin node name\n" \
@@ -66,6 +73,16 @@ CommandLine fork_command =
 		cli_copy_db_getopts,
 		cli_clone);
 
+/* pgcopydb copy db is an alias for pgcopydb clone */
+CommandLine copy__db_command =
+	make_command(
+		"copy-db",
+		"Clone an entire database from source to target",
+		" --source ... --target ... [ --table-jobs ... --index-jobs ... ] ",
+		PGCOPYDB_CLONE_GETOPTS_HELP,
+		cli_copy_db_getopts,
+		cli_clone);
+
 
 CommandLine follow_command =
 	make_command(
@@ -80,6 +97,7 @@ CommandLine follow_command =
 		"  --resume              Allow resuming operations after a failure\n"
 		"  --not-consistent      Allow taking a new snapshot on the source database\n"
 		"  --snapshot            Use snapshot obtained with pg_export_snapshot\n"
+		"  --plugin              Output plugin to use (test_decoding, wal2json)\n" \
 		"  --slot-name           Use this Postgres replication slot name\n"
 		"  --create-slot         Create the replication slot\n"
 		"  --origin              Use this Postgres replication origin node name\n"
@@ -88,6 +106,8 @@ CommandLine follow_command =
 		cli_follow);
 
 
+static void clone_and_follow(CopyDataSpec *copySpecs);
+
 static bool start_clone_process(CopyDataSpec *copySpecs, pid_t *pid);
 
 static bool start_follow_process(CopyDataSpec *copySpecs,
@@ -97,7 +117,6 @@ static bool start_follow_process(CopyDataSpec *copySpecs,
 static bool cli_clone_follow_wait_subprocess(const char *name, pid_t pid);
 
 static bool cloneDB(CopyDataSpec *copySpecs);
-static bool followDB(CopyDataSpec *copySpecs, StreamSpecs *streamSpecs);
 
 
 /*
@@ -107,125 +126,130 @@ void
 cli_clone(int argc, char **argv)
 {
 	CopyDataSpec copySpecs = { 0 };
-	StreamSpecs streamSpecs = { 0 };
 
 	(void) cli_copy_prepare_specs(&copySpecs, DATA_SECTION_ALL);
 
 	/* at the moment this is not covered by cli_copy_prepare_specs() */
 	copySpecs.follow = copyDBoptions.follow;
 
+	/*
+	 * When pgcopydb clone --follow is used, we call the clone_and_follow()
+	 * function which does it all, and just quit.
+	 */
 	if (copySpecs.follow)
 	{
-		if (!stream_init_specs(&streamSpecs,
-							   &(copySpecs.cfPaths.cdc),
-							   copySpecs.source_pguri,
-							   copySpecs.target_pguri,
-							   copyDBoptions.slotName,
-							   copyDBoptions.origin,
-							   copyDBoptions.endpos,
-							   STREAM_MODE_PREFETCH))
-		{
-			/* errors have already been logged */
-			exit(EXIT_CODE_INTERNAL_ERROR);
-		}
+		(void) clone_and_follow(&copySpecs);
+		exit(EXIT_CODE_QUIT);
 	}
 
 	/*
-	 * First, we need to open a snapshot that we're going to re-use in all our
-	 * connections to the source database. When the --snapshot option has been
-	 * used, instead of exporting a new snapshot, we can just re-use it.
+	 * From now on, we know the --follow option has not been used, it's all
+	 * about doing a bare clone operation.
 	 *
-	 * When using PostgreSQL 9.6 logical decoding, we need to create our
-	 * replication slot and fetch the snapshot from that logical replication
-	 * command, it's the only way.
+	 * First, make sure to export a snapshot.
 	 */
-	bool snapshotExported = false;
+	bool exportSnapshot = copydb_should_export_snapshot(&copySpecs);
 
-	if (copySpecs.follow)
-	{
-		TransactionSnapshot *snapshot = &(copySpecs.sourceSnapshot);
-		PGSQL *pgsql = &(snapshot->pgsql);
-
-		if (!pgsql_init(pgsql, copySpecs.source_pguri, PGSQL_CONN_SOURCE))
-		{
-			/* errors have already been logged */
-			exit(EXIT_CODE_SOURCE);
-		}
-
-		if (!pgsql_server_version(pgsql))
-		{
-			/* errors have already been logged */
-			exit(EXIT_CODE_SOURCE);
-		}
-
-		if (pgsql->pgversion_num < 100000)
-		{
-			if (!copydb_create_logical_replication_slot(&copySpecs,
-														streamSpecs.logrep_pguri,
-														streamSpecs.slotName))
-			{
-				/* errors have already been logged */
-				exit(EXIT_CODE_SOURCE);
-			}
-
-			snapshotExported = true;
-		}
-	}
-
-	/*
-	 * Make sure to export a snapshot if that's not been done just above.
-	 */
-	if (!snapshotExported && !copydb_prepare_snapshot(&copySpecs))
+	if (exportSnapshot && !copydb_prepare_snapshot(&copySpecs))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
-	/* just keep track, even though we're not going to use this anymore */
-	snapshotExported = true;
+	pid_t clonePID = -1;
+
+	if (!start_clone_process(&copySpecs, &clonePID))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
+	/* wait until the clone process is finished */
+	bool success = cli_clone_follow_wait_subprocess("clone", clonePID);
+
+	/* close our top-level copy db connection and snapshot */
+	if (exportSnapshot && !copydb_close_snapshot(&copySpecs))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_SOURCE);
+	}
+
+	/* make sure all sub-processes are now finished */
+	bool allExitsAreZero = copydb_wait_for_subprocesses(copySpecs.failFast);
+
+	if (!success || !allExitsAreZero)
+	{
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+}
+
+
+/*
+ * clone_and_follow implements the command: pgcopydb clone --follow
+ */
+static void
+clone_and_follow(CopyDataSpec *copySpecs)
+{
+	StreamSpecs streamSpecs = { 0 };
+
+	/*
+	 * Refrain from logging SQL statements in the apply module, because they
+	 * contain user data. That said, when --trace has been used, bypass that
+	 * privacy feature.
+	 */
+	bool logSQL = log_get_level() <= LOG_TRACE;
+
+	if (!stream_init_specs(&streamSpecs,
+						   &(copySpecs->cfPaths.cdc),
+						   &(copySpecs->connStrings),
+						   &(copyDBoptions.slot),
+						   copyDBoptions.origin,
+						   copyDBoptions.endpos,
+						   STREAM_MODE_CATCHUP,
+						   &(copySpecs->catalog),
+						   copyDBoptions.stdIn,
+						   copyDBoptions.stdOut,
+						   logSQL))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
+	/*
+	 * When using pgcopydb clone --follow --restart we first cleanup the
+	 * previous setup, and that includes dropping the replication slot.
+	 */
+	if (copySpecs->restart)
+	{
+		log_info("Clean-up replication setup, per --restart");
+
+		if (!stream_cleanup_databases(copySpecs,
+									  copyDBoptions.slot.slotName,
+									  copyDBoptions.origin))
+		{
+			/* errors have already been logged */
+			exit(EXIT_CODE_INTERNAL_ERROR);
+		}
+	}
+
+	/*
+	 * First create/export a snapshot for the whole clone --follow operations.
+	 */
+	if (!follow_export_snapshot(copySpecs, &streamSpecs))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_SOURCE);
+	}
 
 	/*
 	 * When --follow has been used, we start two subprocess (clone, follow).
 	 * Before doing that though, we want to make sure it was possible to setup
-	 * the source and target database for Change Data Capture: the wal2json
-	 * logical decoding plugin must be available on the source Postgres
-	 * instance, etc.
+	 * the source and target database for Change Data Capture.
 	 */
-	if (copySpecs.follow)
+	if (!follow_setup_databases(copySpecs, &streamSpecs))
 	{
-		/*
-		 * We want to make sure to use a private PGSQL client connection
-		 * instance when connecting to the source database now, as the main
-		 * connection is currently active holding a snapshot for the whole
-		 * process.
-		 */
-		CopyDataSpec setupSpecs = { 0 };
-		TransactionSnapshot snapshot = { 0 };
-
-		/* copy our structure wholesale */
-		setupSpecs = copySpecs;
-
-		/* ensure we use a new snapshot and connection in setupSpecs */
-		if (!copydb_copy_snapshot(&copySpecs, &snapshot))
-		{
-			/* errors have already been logged */
-			exit(EXIT_CODE_INTERNAL_ERROR);
-		}
-
-		setupSpecs.sourceSnapshot = snapshot;
-
-		/*
-		 * Now create the replication slot and the pgcopydb sentinel table on
-		 * the source database, and the origin (replication progress tracking)
-		 * on the target database.
-		 */
-		if (!stream_setup_databases(&setupSpecs,
-									streamSpecs.slotName,
-									streamSpecs.origin))
-		{
-			/* errors have already been logged */
-			exit(EXIT_CODE_INTERNAL_ERROR);
-		}
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
 	/*
@@ -235,19 +259,16 @@ cli_clone(int argc, char **argv)
 	pid_t clonePID = -1;
 	pid_t followPID = -1;
 
-	if (!start_clone_process(&copySpecs, &clonePID))
+	if (!start_clone_process(copySpecs, &clonePID))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
-	if (copySpecs.follow)
+	if (!start_follow_process(copySpecs, &streamSpecs, &followPID))
 	{
-		if (!start_follow_process(&copySpecs, &streamSpecs, &followPID))
-		{
-			/* errors have already been logged */
-			exit(EXIT_CODE_INTERNAL_ERROR);
-		}
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
 	/* wait until the clone process is finished */
@@ -255,7 +276,7 @@ cli_clone(int argc, char **argv)
 		cli_clone_follow_wait_subprocess("clone", clonePID);
 
 	/* close our top-level copy db connection and snapshot */
-	if (!copydb_close_snapshot(&copySpecs))
+	if (!copydb_close_snapshot(copySpecs))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_SOURCE);
@@ -280,55 +301,35 @@ cli_clone(int argc, char **argv)
 		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
-	/* now wait until the follow process is finished */
-	if (copySpecs.follow)
+	/* now wait until the follow process is finished, if it's been started */
+	if (followPID != -1)
 	{
 		success = success &&
 				  cli_clone_follow_wait_subprocess("follow", followPID);
-
-		/*
-		 * Now is a good time to reset the sequences on the target database to
-		 * match the state they are in at the moment on the source database.
-		 * Postgres logical decoding lacks support for syncing sequences.
-		 *
-		 * This step is implement as if running the following command:
-		 *
-		 *   $ pgcopydb copy sequences --resume --not-consistent
-		 *
-		 * The whole idea is to fetch the "new" current values of the
-		 * sequences, not the ones that were current when the main snapshot was
-		 * exported.
-		 */
-		CopyDataSpec seqSpecs = { 0 };
-
-		/* copy our structure wholesale */
-		seqSpecs = copySpecs;
-
-		/* then force some options such as --resume --not-consistent */
-		seqSpecs.restart = false;
-		seqSpecs.resume = true;
-		seqSpecs.consistent = false;
-		seqSpecs.section = DATA_SECTION_SET_SEQUENCES;
-
-		/* we don't want to re-use any snapshot */
-		TransactionSnapshot snapshot = { 0 };
-
-		seqSpecs.sourceSnapshot = snapshot;
-
-		/* fetch schema information from source catalogs, including filtering */
-		if (!copydb_fetch_schema_and_prepare_specs(&seqSpecs))
-		{
-			/* errors have already been logged */
-			exit(EXIT_CODE_TARGET);
-		}
-
-		if (!copydb_copy_all_sequences(&seqSpecs))
-		{
-			/* errors have already been logged */
-			exit(EXIT_CODE_INTERNAL_ERROR);
-		}
 	}
 
+	/*
+	 * Now is a good time to reset the sequences on the target database to
+	 * match the state they are in at the moment on the source database.
+	 * Postgres logical decoding lacks support for syncing sequences.
+	 *
+	 * This step is implement as if running the following command:
+	 *
+	 *   $ pgcopydb copy sequences --resume --not-consistent
+	 *
+	 * The whole idea is to fetch the "new" current values of the
+	 * sequences, not the ones that were current when the main snapshot was
+	 * exported.
+	 */
+	if (!follow_reset_sequences(copySpecs, &streamSpecs))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_TARGET);
+	}
+
+	/* make sure all sub-processes are now finished */
+	success = success && copydb_wait_for_subprocesses(copySpecs->failFast);
+
 	if (!success)
 	{
 		exit(EXIT_CODE_INTERNAL_ERROR);
@@ -346,58 +347,78 @@ cli_follow(int argc, char **argv)
 
 	(void) cli_copy_prepare_specs(&copySpecs, DATA_SECTION_ALL);
 
+	/*
+	 * Refrain from logging SQL statements in the apply module, because they
+	 * contain user data. That said, when --trace has been used, bypass that
+	 * privacy feature.
+	 */
+	bool logSQL = log_get_level() <= LOG_TRACE;
+
 	StreamSpecs specs = { 0 };
 
 	if (!stream_init_specs(&specs,
 						   &(copySpecs.cfPaths.cdc),
-						   copySpecs.source_pguri,
-						   copySpecs.target_pguri,
-						   copyDBoptions.slotName,
+						   &(copySpecs.connStrings),
+						   &(copyDBoptions.slot),
 						   copyDBoptions.origin,
 						   copyDBoptions.endpos,
-						   STREAM_MODE_PREFETCH))
+						   STREAM_MODE_CATCHUP,
+						   &(copySpecs.catalog),
+						   copyDBoptions.stdIn,
+						   copyDBoptions.stdOut,
+						   logSQL))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
+	/*
+	 * First create/export a snapshot for the whole clone --follow operations.
+	 */
+	if (!follow_export_snapshot(&copySpecs, &specs))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_SOURCE);
+	}
+
 	/*
 	 * First create the replication slot on the source database, and the origin
 	 * (replication progress tracking) on the target database.
 	 */
-	if (!stream_setup_databases(&copySpecs, specs.slotName, specs.origin))
+	if (!follow_setup_databases(&copySpecs, &specs))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
 	/*
-	 * Now remove the possibly still existing stream context files from
-	 * previous round of operations (--resume, etc). We want to make sure that
-	 * the catchup process reads the files created on this connection.
+	 * Before starting the receive, transform, and apply sub-processes, we need
+	 * to set the sentinel endpos to the command line --endpos option, when
+	 * given.
+	 *
+	 * Also fetch the current values from the pgcopydb.sentinel. It might have
+	 * been updated from a previous run of the command, and we might have
+	 * nothing to catch-up to when e.g. the endpos was reached already.
 	 */
-	if (!stream_cleanup_context(&specs))
+	CopyDBSentinel sentinel = { 0 };
+
+	if (!follow_init_sentinel(&specs, &sentinel))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
-	pid_t prefetch = -1;
-	pid_t catchup = -1;
-
-	if (!follow_start_prefetch(&specs, &prefetch))
+	if (sentinel.endpos != InvalidXLogRecPtr &&
+		sentinel.endpos <= sentinel.replay_lsn)
 	{
-		/* errors have already been logged */
-		exit(EXIT_CODE_SOURCE);
+		log_info("Current endpos %X/%X was previously reached at %X/%X",
+				 LSN_FORMAT_ARGS(sentinel.endpos),
+				 LSN_FORMAT_ARGS(sentinel.replay_lsn));
+
+		exit(EXIT_CODE_QUIT);
 	}
 
-	if (!follow_start_catchup(&specs, &catchup))
-	{
-		/* errors have already been logged */
-		exit(EXIT_CODE_TARGET);
-	}
-
-	if (!follow_wait_subprocesses(&specs, prefetch, catchup))
+	if (!follow_main_loop(&copySpecs, &specs))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_INTERNAL_ERROR);
@@ -426,7 +447,7 @@ start_clone_process(CopyDataSpec *copySpecs, pid_t *pid)
 		case 0:
 		{
 			/* child process runs the command */
-			log_debug("Starting the clone sub-process");
+			log_notice("Starting the clone sub-process");
 
 			if (!cloneDB(copySpecs))
 			{
@@ -472,27 +493,15 @@ cloneDB(CopyDataSpec *copySpecs)
 		log_info("STEP 0: copy the source database roles");
 
 		if (!pg_copy_roles(&(copySpecs->pgPaths),
-						   copySpecs->source_pguri,
-						   copySpecs->target_pguri,
-						   copySpecs->dumpPaths.rolesFilename))
+						   &(copySpecs->connStrings),
+						   copySpecs->dumpPaths.rolesFilename,
+						   copySpecs->noRolesPasswords))
 		{
 			/* errors have already been logged */
 			return false;
 		}
 	}
 
-	log_info("STEP 1: dump the source database schema (pre/post data)");
-
-	(void) summary_set_current_time(timings, TIMING_STEP_BEFORE_SCHEMA_DUMP);
-
-	if (!copydb_dump_source_schema(copySpecs,
-								   copySpecs->sourceSnapshot.snapshot,
-								   PG_DUMP_SECTION_SCHEMA))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
 	/* make sure that we have our own process local connection */
 	TransactionSnapshot snapshot = { 0 };
 
@@ -512,8 +521,7 @@ cloneDB(CopyDataSpec *copySpecs)
 	}
 
 	/* fetch schema information from source catalogs, including filtering */
-
-	log_info("STEP 2: fetch source database tables, indexes, and sequences");
+	log_info("STEP 1: fetch source database tables, indexes, and sequences");
 
 	(void) summary_set_current_time(timings, TIMING_STEP_BEFORE_SCHEMA_FETCH);
 
@@ -530,10 +538,22 @@ cloneDB(CopyDataSpec *copySpecs)
 		return false;
 	}
 
-	(void) summary_set_current_time(timings, TIMING_STEP_BEFORE_PREPARE_SCHEMA);
+	log_info("STEP 2: dump the source database schema (pre/post data)");
+
+	(void) summary_set_current_time(timings, TIMING_STEP_BEFORE_SCHEMA_DUMP);
+
+	if (!copydb_dump_source_schema(copySpecs,
+								   copySpecs->sourceSnapshot.snapshot,
+								   PG_DUMP_SECTION_SCHEMA))
+	{
+		/* errors have already been logged */
+		return false;
+	}
 
 	log_info("STEP 3: restore the pre-data section to the target database");
 
+	(void) summary_set_current_time(timings, TIMING_STEP_BEFORE_PREPARE_SCHEMA);
+
 	if (!copydb_target_prepare_schema(copySpecs))
 	{
 		/* errors have already been logged */
@@ -542,11 +562,7 @@ cloneDB(CopyDataSpec *copySpecs)
 
 	(void) summary_set_current_time(timings, TIMING_STEP_AFTER_PREPARE_SCHEMA);
 
-	log_info("STEP 4: copy data from source to target in %d sub-processes",
-			 copySpecs->tableJobs);
-
-	/* STEPs 5, 6, 7, 8, and 9 are printed when starting the sub-processes */
-
+	/* STEPs 4, 5, 6, 7, 8, and 9 are printed when starting the sub-processes */
 	if (!copydb_copy_all_table_data(copySpecs))
 	{
 		/* errors have already been logged */
@@ -576,7 +592,9 @@ cloneDB(CopyDataSpec *copySpecs)
 
 		log_info("Updating the pgcopydb.sentinel to enable applying changes");
 
-		if (!pgsql_init(&pgsql, copySpecs->source_pguri, PGSQL_CONN_SOURCE))
+		if (!pgsql_init(&pgsql,
+						copySpecs->connStrings.source_pguri,
+						PGSQL_CONN_SOURCE))
 		{
 			/* errors have already been logged */
 			return false;
@@ -606,6 +624,33 @@ static bool
 start_follow_process(CopyDataSpec *copySpecs, StreamSpecs *streamSpecs,
 					 pid_t *pid)
 {
+	/*
+	 * Before starting the receive, transform, and apply sub-processes, we need
+	 * to set the sentinel endpos to the command line --endpos option, when
+	 * given.
+	 *
+	 * Also fetch the current values from the pgcopydb.sentinel. It might have
+	 * been updated from a previous run of the command, and we might have
+	 * nothing to catch-up to when e.g. the endpos was reached already.
+	 */
+	CopyDBSentinel sentinel = { 0 };
+
+	if (!follow_init_sentinel(streamSpecs, &sentinel))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+	if (sentinel.endpos != InvalidXLogRecPtr &&
+		sentinel.endpos <= sentinel.replay_lsn)
+	{
+		log_info("Current endpos %X/%X was previously reached at %X/%X",
+				 LSN_FORMAT_ARGS(sentinel.endpos),
+				 LSN_FORMAT_ARGS(sentinel.replay_lsn));
+
+		return true;
+	}
+
 	/* now we can fork a sub-process to transform the current file */
 	pid_t fpid = fork();
 
@@ -620,9 +665,9 @@ start_follow_process(CopyDataSpec *copySpecs, StreamSpecs *streamSpecs,
 		case 0:
 		{
 			/* child process runs the command */
-			log_debug("Starting the follow sub-process");
+			log_notice("Starting the follow sub-process");
 
-			if (!followDB(copySpecs, streamSpecs))
+			if (!follow_main_loop(copySpecs, streamSpecs))
 			{
 				/* errors have already been logged */
 				exit(EXIT_CODE_INTERNAL_ERROR);
@@ -643,61 +688,6 @@ start_follow_process(CopyDataSpec *copySpecs, StreamSpecs *streamSpecs,
 }
 
 
-/*
- * followDB implements a logical decoding client for streaming changes from the
- * source database into the target database. The stream_setup_databases() must
- * have been called already so that the replication slot using wal2json is
- * ready, the pgcopydb.sentinel table exists, and the target database
- * replication origin has been created too.
- */
-static bool
-followDB(CopyDataSpec *copySpecs, StreamSpecs *streamSpecs)
-{
-	/*
-	 * Remove the possibly still existing stream context files from
-	 * previous round of operations (--resume, etc). We want to make sure
-	 * that the catchup process reads the files created on this connection.
-	 */
-	if (!stream_cleanup_context(streamSpecs))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	pid_t prefetch = -1;
-	pid_t catchup = -1;
-
-	if (!follow_start_prefetch(streamSpecs, &prefetch))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	/*
-	 * Second, start the catchup process.
-	 */
-	if (!follow_start_catchup(streamSpecs, &catchup))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	/*
-	 * Finally wait until the process are finished.
-	 *
-	 * This happens when the sentinel endpos is set, typically using the
-	 * command: pgcopydb stream sentinel set endpos --current.
-	 */
-	if (!follow_wait_subprocesses(streamSpecs, prefetch, catchup))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	return true;
-}
-
-
 /*
  * cli_clone_follow_wait_subprocesses waits until both sub-processes are
  * finished.
@@ -708,6 +698,12 @@ cli_clone_follow_wait_subprocess(const char *name, pid_t pid)
 	bool exited = false;
 	int returnCode = -1;
 
+	if (pid < 0)
+	{
+		log_error("BUG: cli_clone_follow_wait_subprocess(%s, %d)", name, pid);
+		return false;
+	}
+
 	while (!exited)
 	{
 		if (!follow_wait_pid(pid, &exited, &returnCode))
