@@ -26,7 +26,11 @@
 ListDBOptions listDBoptions = { 0 };
 
 static int cli_list_db_getopts(int argc, char **argv);
+static void cli_list_databases(int argc, char **argv);
 static void cli_list_extensions(int argc, char **argv);
+static void cli_list_extension_versions(int argc, char **argv);
+static void cli_list_extension_requirements(int argc, char **argv);
+static void cli_list_collations(int argc, char **argv);
 static void cli_list_tables(int argc, char **argv);
 static void cli_list_table_parts(int argc, char **argv);
 static void cli_list_sequences(int argc, char **argv);
@@ -35,15 +39,39 @@ static void cli_list_depends(int argc, char **argv);
 static void cli_list_schema(int argc, char **argv);
 static void cli_list_progress(int argc, char **argv);
 
+static bool copydb_init_specs_from_listdboptions(CopyDBOptions *options,
+												 ListDBOptions *listDBoptions);
+
+static CommandLine list_catalogs_command =
+	make_command(
+		"databases",
+		"List databases",
+		" --source ... ",
+		"  --source            Postgres URI to the source database\n",
+		cli_list_db_getopts,
+		cli_list_databases);
+
 static CommandLine list_extensions_command =
 	make_command(
 		"extensions",
 		"List all the source extensions to copy",
 		" --source ... ",
-		"  --source            Postgres URI to the source database\n",
+		"  --source              Postgres URI to the source database\n"
+		"  --json                Format the output using JSON\n"
+		"  --available-versions  List available extension versions\n"
+		"  --requirements        List extensions requirements\n",
 		cli_list_db_getopts,
 		cli_list_extensions);
 
+static CommandLine list_collations_command =
+	make_command(
+		"collations",
+		"List all the source collations to copy",
+		" --source ... ",
+		"  --source            Postgres URI to the source database\n",
+		cli_list_db_getopts,
+		cli_list_collations);
+
 static CommandLine list_tables_command =
 	make_command(
 		"tables",
@@ -51,6 +79,8 @@ static CommandLine list_tables_command =
 		" --source ... ",
 		"  --source            Postgres URI to the source database\n"
 		"  --filter <filename> Use the filters defined in <filename>\n"
+		"  --cache             Cache table size in relation pgcopydb.pgcopydb_table_size\n"
+		"  --drop-cache        Drop relation pgcopydb.pgcopydb_table_size\n"
 		"  --list-skipped      List only tables that are setup to be skipped\n"
 		"  --without-pkey      List only tables that have no primary key\n",
 		cli_list_db_getopts,
@@ -121,13 +151,17 @@ static CommandLine list_progress_command =
 		"List the progress",
 		" --source ... ",
 		"  --source  Postgres URI to the source database\n"
-		"  --json    Format the output using JSON\n",
+		"  --summary List the summary, requires --json\n"
+		"  --json    Format the output using JSON\n"
+		"  --dir     Work directory to use\n",
 		cli_list_db_getopts,
 		cli_list_progress);
 
 
 static CommandLine *list_subcommands[] = {
+	&list_catalogs_command,
 	&list_extensions_command,
+	&list_collations_command,
 	&list_tables_command,
 	&list_table_parts_command,
 	&list_sequences_command,
@@ -143,6 +177,32 @@ CommandLine list_commands =
 					 "List database objects from a Postgres instance",
 					 NULL, NULL, NULL, list_subcommands);
 
+
+/*
+ * cli_list_getenv reads from the environment variables and fills-in the
+ * command line options.
+ */
+static int
+cli_list_getenv(ListDBOptions *options)
+{
+	int errors = 0;
+
+	if (!cli_copydb_getenv_source_pguri(&(options->connStrings.source_pguri)))
+	{
+		/* errors have already been logged */
+		++errors;
+	}
+
+	if (!cli_copydb_getenv_split(&(options->splitTablesLargerThan)))
+	{
+		/* errors have already been logged */
+		++errors;
+	}
+
+	return errors == 0;
+}
+
+
 /*
  * cli_list_db_getopts parses the CLI options for the `list db` command.
  */
@@ -155,6 +215,7 @@ cli_list_db_getopts(int argc, char **argv)
 
 	static struct option long_options[] = {
 		{ "source", required_argument, NULL, 'S' },
+		{ "dir", required_argument, NULL, 'D' },
 		{ "schema-name", required_argument, NULL, 's' },
 		{ "table-name", required_argument, NULL, 't' },
 		{ "filter", required_argument, NULL, 'F' },
@@ -163,11 +224,17 @@ cli_list_db_getopts(int argc, char **argv)
 		{ "without-pkey", no_argument, NULL, 'P' },
 		{ "split-tables-larger-than", required_argument, NULL, 'L' },
 		{ "split-at", required_argument, NULL, 'L' },
+		{ "cache", no_argument, NULL, 'c' },
+		{ "drop-cache", no_argument, NULL, 'C' },
+		{ "summary", no_argument, NULL, 'y' },
+		{ "available-versions", no_argument, NULL, 'a' },
+		{ "requirements", no_argument, NULL, 'r' },
 		{ "json", no_argument, NULL, 'J' },
 		{ "version", no_argument, NULL, 'V' },
 		{ "debug", no_argument, NULL, 'd' },
 		{ "trace", no_argument, NULL, 'z' },
 		{ "verbose", no_argument, NULL, 'v' },
+		{ "notice", no_argument, NULL, 'v' },
 		{ "quiet", no_argument, NULL, 'q' },
 		{ "help", no_argument, NULL, 'h' },
 		{ NULL, 0, NULL, 0 }
@@ -175,7 +242,14 @@ cli_list_db_getopts(int argc, char **argv)
 
 	optind = 0;
 
-	while ((c = getopt_long(argc, argv, "S:T:j:s:t:PL:JVvdzqh",
+	/* read values from the environment */
+	if (!cli_list_getenv(&options))
+	{
+		log_fatal("Failed to read default values from the environment");
+		exit(EXIT_CODE_BAD_ARGS);
+	}
+
+	while ((c = getopt_long(argc, argv, "S:T:D:j:s:t:PL:cCJVvdzqh",
 							long_options, &option_index)) != -1)
 	{
 		switch (c)
@@ -188,8 +262,8 @@ cli_list_db_getopts(int argc, char **argv)
 							  "see above for details.");
 					exit(EXIT_CODE_BAD_ARGS);
 				}
-				strlcpy(options.source_pguri, optarg, MAXCONNINFO);
-				log_trace("--source %s", options.source_pguri);
+				options.connStrings.source_pguri = pg_strdup(optarg);
+				log_trace("--source %s", options.connStrings.source_pguri);
 				break;
 			}
 
@@ -200,6 +274,13 @@ cli_list_db_getopts(int argc, char **argv)
 				break;
 			}
 
+			case 'D':
+			{
+				strlcpy(options.dir, optarg, MAXPGPATH);
+				log_trace("--dir %s", options.dir);
+				break;
+			}
+
 			case 't':
 			{
 				strlcpy(options.table_name, optarg, NAMEDATALEN);
@@ -239,9 +320,9 @@ cli_list_db_getopts(int argc, char **argv)
 			{
 				if (!cli_parse_bytes_pretty(
 						optarg,
-						&options.splitTablesLargerThan,
-						(char *) &options.splitTablesLargerThanPretty,
-						sizeof(options.splitTablesLargerThanPretty)))
+						&(options.splitTablesLargerThan.bytes),
+						(char *) &(options.splitTablesLargerThan.bytesPretty),
+						sizeof(options.splitTablesLargerThan.bytesPretty)))
 				{
 					log_fatal("Failed to parse --split-tables-larger-than: \"%s\"",
 							  optarg);
@@ -249,8 +330,55 @@ cli_list_db_getopts(int argc, char **argv)
 				}
 
 				log_trace("--split-tables-larger-than %s (%lld)",
-						  options.splitTablesLargerThanPretty,
-						  (long long) options.splitTablesLargerThan);
+						  options.splitTablesLargerThan.bytesPretty,
+						  (long long) options.splitTablesLargerThan.bytes);
+				break;
+			}
+
+			case 'c':
+			{
+				if (options.dropCache)
+				{
+					log_fatal("Please choose either --cache or --drop-cache");
+					++errors;
+				}
+
+				options.cache = true;
+				log_trace("--cache");
+				break;
+			}
+
+			case 'C':
+			{
+				if (options.cache)
+				{
+					log_fatal("Please choose either --cache or --drop-cache");
+					++errors;
+				}
+
+				options.dropCache = true;
+				log_trace("--drop-cache");
+				break;
+			}
+
+			case 'y':
+			{
+				options.summary = true;
+				log_trace("--summary");
+				break;
+			}
+
+			case 'a':
+			{
+				options.availableVersions = true;
+				log_trace("--available-versions");
+				break;
+			}
+
+			case 'r':
+			{
+				options.requirements = true;
+				log_trace("--requirements");
 				break;
 			}
 
@@ -280,6 +408,12 @@ cli_list_db_getopts(int argc, char **argv)
 					}
 
 					case 2:
+					{
+						log_set_level(LOG_SQL);
+						break;
+					}
+
+					case 3:
 					{
 						log_set_level(LOG_DEBUG);
 						break;
@@ -296,14 +430,14 @@ cli_list_db_getopts(int argc, char **argv)
 
 			case 'd':
 			{
-				verboseCount = 2;
+				verboseCount = 3;
 				log_set_level(LOG_DEBUG);
 				break;
 			}
 
 			case 'z':
 			{
-				verboseCount = 3;
+				verboseCount = 4;
 				log_set_level(LOG_TRACE);
 				break;
 			}
@@ -328,58 +462,25 @@ cli_list_db_getopts(int argc, char **argv)
 		}
 	}
 
-	/* list commands support the source URI environment variable */
-	if (IS_EMPTY_STRING_BUFFER(options.source_pguri))
-	{
-		if (env_exists(PGCOPYDB_SOURCE_PGURI))
-		{
-			if (!get_env_copy(PGCOPYDB_SOURCE_PGURI,
-							  options.source_pguri,
-							  sizeof(options.source_pguri)))
-			{
-				/* errors have already been logged */
-				++errors;
-			}
-		}
-	}
-
-	if (IS_EMPTY_STRING_BUFFER(options.source_pguri))
+	if (options.connStrings.source_pguri == NULL)
 	{
 		log_fatal("Option --source is mandatory");
 		++errors;
 	}
 
+	/* prepare safe versions of the connection strings (without password) */
+	if (!cli_prepare_pguris(&(options.connStrings)))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
 	if (options.listSkipped && IS_EMPTY_STRING_BUFFER(options.filterFileName))
 	{
 		log_fatal("Option --list-skipped requires using option --filters");
 		++errors;
 	}
 
-	if (env_exists(PGCOPYDB_SPLIT_TABLES_LARGER_THAN))
-	{
-		char bytes[BUFSIZE] = { 0 };
-
-		if (get_env_copy(PGCOPYDB_SPLIT_TABLES_LARGER_THAN, bytes, sizeof(bytes)))
-		{
-			if (!cli_parse_bytes_pretty(
-					bytes,
-					&options.splitTablesLargerThan,
-					(char *) &options.splitTablesLargerThanPretty,
-					sizeof(options.splitTablesLargerThanPretty)))
-			{
-				log_fatal("Failed to parse PGCOPYDB_SPLIT_TABLES_LARGER_THAN: "
-						  " \"%s\"",
-						  bytes);
-				++errors;
-			}
-		}
-		else
-		{
-			/* errors have already been logged */
-			++errors;
-		}
-	}
-
 	if (errors > 0)
 	{
 		exit(EXIT_CODE_BAD_ARGS);
@@ -392,16 +493,81 @@ cli_list_db_getopts(int argc, char **argv)
 }
 
 
+/*
+ * cli_list_databases implements the command: pgcopydb list databases
+ */
+static void
+cli_list_databases(int argc, char **argv)
+{
+	PGSQL pgsql = { 0 };
+	SourceDatabaseArray databaseArray = { 0, NULL };
+
+	ConnStrings *dsn = &(listDBoptions.connStrings);
+
+	if (!pgsql_init(&pgsql, dsn->source_pguri, PGSQL_CONN_SOURCE))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_SOURCE);
+	}
+
+	if (!schema_list_databases(&pgsql, &databaseArray))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
+	log_info("Fetched information for %d databases", databaseArray.count);
+
+	fformat(stdout, "%10s | %20s | %20s\n",
+			"OID",
+			"Database Name",
+			"On-disk size");
+
+	fformat(stdout, "%10s-+-%20s-+-%20s\n",
+			"----------",
+			"--------------------",
+			"--------------------");
+
+	for (int i = 0; i < databaseArray.count; i++)
+	{
+		SourceDatabase *cat = &(databaseArray.array[i]);
+
+		fformat(stdout, "%10u | %20s | %20s\n",
+				cat->oid,
+				cat->datname,
+				cat->bytesPretty);
+	}
+
+	fformat(stdout, "\n");
+}
+
+
 /*
  * cli_list_extensions implements the command: pgcopydb list extensions
  */
 static void
 cli_list_extensions(int argc, char **argv)
 {
+	/* --available-versions is implemented as its own command */
+	if (listDBoptions.availableVersions)
+	{
+		(void) cli_list_extension_versions(argc, argv);
+		exit(EXIT_CODE_QUIT);
+	}
+
+	/* --requirements is implemented as its own command */
+	if (listDBoptions.requirements)
+	{
+		(void) cli_list_extension_requirements(argc, argv);
+		exit(EXIT_CODE_QUIT);
+	}
+
 	PGSQL pgsql = { 0 };
 	SourceExtensionArray extensionArray = { 0, NULL };
 
-	if (!pgsql_init(&pgsql, listDBoptions.source_pguri, PGSQL_CONN_SOURCE))
+	ConnStrings *dsn = &(listDBoptions.connStrings);
+
+	if (!pgsql_init(&pgsql, dsn->source_pguri, PGSQL_CONN_SOURCE))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_SOURCE);
@@ -415,41 +581,297 @@ cli_list_extensions(int argc, char **argv)
 
 	log_info("Fetched information for %d extensions", extensionArray.count);
 
-	fformat(stdout, "%10s | %20s | %20s | %10s | %s\n",
-			"OID",
-			"Name",
-			"Schema",
-			"Count",
-			"Config");
-
-	fformat(stdout, "%10s-+-%20s-+-%20s-+-%10s-+-%10s\n",
-			"----------",
-			"--------------------",
-			"--------------------",
-			"----------",
-			"----------");
-
-	for (int i = 0; i < extensionArray.count; i++)
+	if (outputJSON)
 	{
-		SourceExtension *ext = &(extensionArray.array[i]);
+		JSON_Value *js = json_value_init_array();
+		JSON_Array *jsArray = json_value_get_array(js);
 
-		char config[BUFSIZE] = { 0 };
-
-		for (int c = 0; c < ext->config.count; c++)
+		for (int i = 0; i < extensionArray.count; i++)
 		{
-			sformat(config, sizeof(config), "%s%s\"%s\".\"%s\"",
-					config,
-					c == 0 ? "" : ",",
-					ext->config.array[c].nspname,
-					ext->config.array[c].relname);
+			SourceExtension *ext = &(extensionArray.array[i]);
+
+			JSON_Value *jsExt = json_value_init_object();
+			JSON_Object *jsObj = json_value_get_object(jsExt);
+
+			json_object_set_number(jsObj, "oid", ext->oid);
+			json_object_set_string(jsObj, "name", ext->extname);
+			json_object_set_string(jsObj, "schema", ext->extnamespace);
+
+			JSON_Value *jsConfig = json_value_init_array();
+			JSON_Array *jsConfigArray = json_value_get_array(jsConfig);
+
+			for (int c = 0; c < ext->config.count; c++)
+			{
+				JSON_Value *jsConf = json_value_init_object();
+				JSON_Object *jsConfObj = json_value_get_object(jsConf);
+
+				json_object_set_string(jsConfObj,
+									   "schema", ext->config.array[c].nspname);
+
+				json_object_set_string(jsConfObj,
+									   "name", ext->config.array[c].relname);
+
+				json_array_append_value(jsConfigArray, jsConf);
+			}
+
+			json_object_set_value(jsObj, "config", jsConfig);
+
+			json_array_append_value(jsArray, jsExt);
 		}
 
-		fformat(stdout, "%10u | %20s | %20s | %10d | %s\n",
-				ext->oid,
-				ext->extname,
-				ext->extnamespace,
-				ext->config.count,
-				config);
+		char *serialized_string = json_serialize_to_string_pretty(js);
+
+		fformat(stdout, "%s\n", serialized_string);
+
+		json_free_serialized_string(serialized_string);
+		json_value_free(js);
+	}
+	else
+	{
+		fformat(stdout, "%10s | %20s | %20s | %10s | %s\n",
+				"OID",
+				"Name",
+				"Schema",
+				"Count",
+				"Config");
+
+		fformat(stdout, "%10s-+-%20s-+-%20s-+-%10s-+-%10s\n",
+				"----------",
+				"--------------------",
+				"--------------------",
+				"----------",
+				"----------");
+
+		for (int i = 0; i < extensionArray.count; i++)
+		{
+			SourceExtension *ext = &(extensionArray.array[i]);
+
+			char config[BUFSIZE] = { 0 };
+
+			for (int c = 0; c < ext->config.count; c++)
+			{
+				sformat(config, sizeof(config), "%s%s\"%s\".\"%s\"",
+						config,
+						c == 0 ? "" : ",",
+						ext->config.array[c].nspname,
+						ext->config.array[c].relname);
+			}
+
+			fformat(stdout, "%10u | %20s | %20s | %10d | %s\n",
+					ext->oid,
+					ext->extname,
+					ext->extnamespace,
+					ext->config.count,
+					config);
+		}
+
+		fformat(stdout, "\n");
+	}
+}
+
+
+/*
+ * cli_list_extension_versions implements the command:
+ *
+ *   pgcopydb list extensions --available-versions
+ */
+static void
+cli_list_extension_versions(int argc, char **argv)
+{
+	PGSQL pgsql = { 0 };
+	ExtensionsVersionsArray evArray = { 0, NULL };
+
+	ConnStrings *dsn = &(listDBoptions.connStrings);
+
+	if (!pgsql_init(&pgsql, dsn->source_pguri, PGSQL_CONN_SOURCE))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_SOURCE);
+	}
+
+	if (!schema_list_ext_versions(&pgsql, &evArray))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
+	log_info("Fetched information for %d extensions", evArray.count);
+
+	if (outputJSON)
+	{
+		JSON_Value *js = json_value_init_array();
+		JSON_Array *jsArray = json_value_get_array(js);
+
+		for (int i = 0; i < evArray.count; i++)
+		{
+			ExtensionsVersions *ev = &(evArray.array[i]);
+
+			JSON_Value *jsExtVersion = json_value_init_object();
+			JSON_Object *jsEVObj = json_value_get_object(jsExtVersion);
+
+			json_object_set_string(jsEVObj, "name", ev->name);
+			json_object_set_value(jsEVObj, "versions", ev->json);
+
+			/* add the JSON object to the array */
+			json_array_append_value(jsArray, jsExtVersion);
+		}
+
+		char *serialized_string = json_serialize_to_string_pretty(js);
+
+		fformat(stdout, "%s\n", serialized_string);
+
+		json_free_serialized_string(serialized_string);
+		json_value_free(js);
+	}
+	else
+	{
+		fformat(stdout, "%20s | %20s | %s\n",
+				"Name",
+				"Default",
+				"Available");
+
+		fformat(stdout, "%20s-+-%20s-+-%20s\n",
+				"--------------------",
+				"--------------------",
+				"--------------------");
+
+		for (int i = 0; i < evArray.count; i++)
+		{
+			ExtensionsVersions *ev = &(evArray.array[i]);
+
+			char *strArray = json_serialize_to_string(ev->json);
+
+			fformat(stdout, "%20s | %20s | %s\n",
+					ev->name,
+					ev->defaultVersion,
+					strArray);
+
+			json_free_serialized_string(strArray);
+		}
+
+		fformat(stdout, "\n");
+	}
+}
+
+
+/*
+ * cli_list_extension_requirements implements the command:
+ *
+ *   pgcopydb list extensions --requirements --json
+ */
+static void
+cli_list_extension_requirements(int argc, char **argv)
+{
+	PGSQL pgsql = { 0 };
+	ExtensionsVersionsArray evArray = { 0, NULL };
+
+	ConnStrings *dsn = &(listDBoptions.connStrings);
+
+	if (!pgsql_init(&pgsql, dsn->source_pguri, PGSQL_CONN_SOURCE))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_SOURCE);
+	}
+
+	if (!schema_list_ext_versions(&pgsql, &evArray))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
+	log_info("Fetched information for %d extensions", evArray.count);
+
+	if (outputJSON)
+	{
+		JSON_Value *js = json_value_init_array();
+		JSON_Array *jsArray = json_value_get_array(js);
+
+		for (int i = 0; i < evArray.count; i++)
+		{
+			ExtensionsVersions *ev = &(evArray.array[i]);
+
+			JSON_Value *jsExtVersion = json_value_init_object();
+			JSON_Object *jsEVObj = json_value_get_object(jsExtVersion);
+
+			json_object_set_string(jsEVObj, "name", ev->name);
+			json_object_set_string(jsEVObj, "version", ev->defaultVersion);
+
+			/* add the JSON object to the array */
+			json_array_append_value(jsArray, jsExtVersion);
+		}
+
+		char *serialized_string = json_serialize_to_string_pretty(js);
+
+		fformat(stdout, "%s\n", serialized_string);
+
+		json_free_serialized_string(serialized_string);
+		json_value_free(js);
+	}
+	else
+	{
+		fformat(stdout, "%30s | %s\n", "Name", "Version");
+
+		fformat(stdout, "%30s-+-%20s\n",
+				"------------------------------",
+				"--------------------");
+
+		for (int i = 0; i < evArray.count; i++)
+		{
+			ExtensionsVersions *ev = &(evArray.array[i]);
+
+			fformat(stdout, "%30s | %s\n",
+					ev->name,
+					ev->defaultVersion);
+		}
+
+		fformat(stdout, "\n");
+	}
+}
+
+
+/*
+ * cli_list_collations implements the command: pgcopydb list collations
+ */
+static void
+cli_list_collations(int argc, char **argv)
+{
+	PGSQL pgsql = { 0 };
+	SourceCollationArray collationArray = { 0, NULL };
+
+	ConnStrings *dsn = &(listDBoptions.connStrings);
+
+	if (!pgsql_init(&pgsql, dsn->source_pguri, PGSQL_CONN_SOURCE))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_SOURCE);
+	}
+
+	if (!schema_list_collations(&pgsql, &collationArray))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
+	log_info("Fetched information for %d collations", collationArray.count);
+
+	fformat(stdout, "%10s | %20s | %20s \n",
+			"OID",
+			"Name",
+			"Object name");
+
+	fformat(stdout, "%10s-+-%20s-+-%20s\n",
+			"----------",
+			"--------------------",
+			"--------------------");
+
+	for (int i = 0; i < collationArray.count; i++)
+	{
+		SourceCollation *coll = &(collationArray.array[i]);
+
+		fformat(stdout, "%10u | %20s | %20s \n",
+				coll->oid,
+				coll->collname,
+				coll->desc);
 	}
 
 	fformat(stdout, "\n");
@@ -492,12 +914,63 @@ cli_list_tables(int argc, char **argv)
 		}
 	}
 
-	if (!pgsql_init(&pgsql, listDBoptions.source_pguri, PGSQL_CONN_SOURCE))
+	ConnStrings *dsn = &(listDBoptions.connStrings);
+
+	if (!pgsql_init(&pgsql, dsn->source_pguri, PGSQL_CONN_SOURCE))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_SOURCE);
 	}
 
+	if (listDBoptions.dropCache)
+	{
+		log_info("Dropping cache table pgcopydb.pgcopydb_table_size");
+		if (!schema_drop_pgcopydb_table_size(&pgsql))
+		{
+			exit(EXIT_CODE_SOURCE);
+		}
+
+		exit(EXIT_CODE_QUIT);
+	}
+
+	if (!pgsql_begin(&pgsql))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_SOURCE);
+	}
+
+	bool hasDBCreatePrivilege;
+	bool hasDBTempPrivilege;
+
+	/* check if we have needed privileges here */
+	if (!schema_query_privileges(&pgsql,
+								 &hasDBCreatePrivilege,
+								 &hasDBTempPrivilege))
+	{
+		log_error("Failed to query database privileges, see above for details");
+		exit(EXIT_CODE_SOURCE);
+	}
+
+	if (!pgsql_prepend_search_path(&pgsql, "pgcopydb"))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_SOURCE);
+	}
+
+	bool createdTableSizeTable = false;
+	bool dropCache = listDBoptions.cache;
+
+	if (!schema_prepare_pgcopydb_table_size(&pgsql,
+											&filters,
+											hasDBCreatePrivilege,
+											listDBoptions.cache,
+											dropCache,
+											&createdTableSizeTable))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
 	if (listDBoptions.noPKey)
 	{
 		log_info("Listing tables without primary key in source database");
@@ -521,6 +994,12 @@ cli_list_tables(int argc, char **argv)
 		}
 	}
 
+	if (!pgsql_commit(&pgsql))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_SOURCE);
+	}
+
 	/* compute total bytes and total reltuples, pretty print them */
 	uint64_t totalBytes = 0;
 	uint64_t totalTuples = 0;
@@ -565,6 +1044,15 @@ cli_list_tables(int argc, char **argv)
 	}
 
 	fformat(stdout, "\n");
+
+	if (createdTableSizeTable && !listDBoptions.cache)
+	{
+		if (!schema_drop_pgcopydb_table_size(&pgsql))
+		{
+			/* errors have already been logged */
+			exit(EXIT_CODE_INTERNAL_ERROR);
+		}
+	}
 }
 
 
@@ -575,7 +1063,6 @@ static void
 cli_list_table_parts(int argc, char **argv)
 {
 	PGSQL pgsql = { 0 };
-	char scrubbedSourceURI[MAXCONNINFO] = { 0 };
 
 	if (IS_EMPTY_STRING_BUFFER(listDBoptions.table_name))
 	{
@@ -589,15 +1076,20 @@ cli_list_table_parts(int argc, char **argv)
 		strlcpy(listDBoptions.schema_name, "public", NAMEDATALEN);
 	}
 
-	(void) parse_and_scrub_connection_string(listDBoptions.source_pguri,
-											 scrubbedSourceURI);
+	ConnStrings *dsn = &(listDBoptions.connStrings);
 
 	log_info("Listing COPY partitions for table \"%s\".\"%s\" in \"%s\"",
 			 listDBoptions.schema_name,
 			 listDBoptions.table_name,
-			 scrubbedSourceURI);
+			 dsn->safeSourcePGURI.pguri);
 
-	if (!pgsql_init(&pgsql, listDBoptions.source_pguri, PGSQL_CONN_SOURCE))
+	if (!pgsql_init(&pgsql, dsn->source_pguri, PGSQL_CONN_SOURCE))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_SOURCE);
+	}
+
+	if (!pgsql_begin(&pgsql))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_SOURCE);
@@ -624,6 +1116,27 @@ cli_list_table_parts(int argc, char **argv)
 		}
 	};
 
+	/* just don't query for privileges, assume read-only */
+	bool hasDBCreatePrivilege = false;
+	bool createdTableSizeTable = false;
+
+	if (!pgsql_prepend_search_path(&pgsql, "pgcopydb"))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_SOURCE);
+	}
+
+	if (!schema_prepare_pgcopydb_table_size(&pgsql,
+											&filter,
+											hasDBCreatePrivilege,
+											false, /* cache */
+											false, /* dropCache */
+											&createdTableSizeTable))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
+
 	SourceTableArray tableArray = { 0 };
 
 	if (!schema_list_ordinary_tables(&pgsql, &filter, &tableArray))
@@ -656,11 +1169,17 @@ cli_list_table_parts(int argc, char **argv)
 	}
 
 	if (!schema_list_partitions(&pgsql, table,
-								listDBoptions.splitTablesLargerThan))
+								listDBoptions.splitTablesLargerThan.bytes))
 	{
 		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
+	if (!pgsql_commit(&pgsql))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_SOURCE);
+	}
+
 	if (table->partsArray.count <= 1)
 	{
 		log_info("Table \"%s\".\"%s\" (%s) will not be split",
@@ -743,7 +1262,9 @@ cli_list_sequences(int argc, char **argv)
 		}
 	}
 
-	if (!pgsql_init(&pgsql, listDBoptions.source_pguri, PGSQL_CONN_SOURCE))
+	ConnStrings *dsn = &(listDBoptions.connStrings);
+
+	if (!pgsql_init(&pgsql, dsn->source_pguri, PGSQL_CONN_SOURCE))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_SOURCE);
@@ -757,20 +1278,27 @@ cli_list_sequences(int argc, char **argv)
 
 	log_info("Fetched information for %d sequences", sequenceArray.count);
 
-	fformat(stdout, "%8s | %20s | %30s\n",
-			"OID", "Schema Name", "Sequence Name");
+	fformat(stdout, "%8s | %20s | %30s | %10s | %10s | %10s \n",
+			"OID", "Schema Name", "Sequence Name",
+			"Owned By", "attrelid", "attroid");
 
-	fformat(stdout, "%8s-+-%20s-+-%30s\n",
+	fformat(stdout, "%8s-+-%20s-+-%30s-+-%10s-+-%10s-+-%10s\n",
 			"--------",
 			"--------------------",
-			"------------------------------");
+			"------------------------------",
+			"----------",
+			"----------",
+			"----------");
 
 	for (int i = 0; i < sequenceArray.count; i++)
 	{
-		fformat(stdout, "%8d | %20s | %30s\n",
+		fformat(stdout, "%8d | %20s | %30s | %10d | %10d | %10d\n",
 				sequenceArray.array[i].oid,
 				sequenceArray.array[i].nspname,
-				sequenceArray.array[i].relname);
+				sequenceArray.array[i].relname,
+				sequenceArray.array[i].ownedby,
+				sequenceArray.array[i].attrelid,
+				sequenceArray.array[i].attroid);
 	}
 
 	fformat(stdout, "\n");
@@ -786,14 +1314,11 @@ cli_list_indexes(int argc, char **argv)
 	PGSQL pgsql = { 0 };
 	SourceIndexArray indexArray = { 0, NULL };
 
-	char scrubbedSourceURI[MAXCONNINFO] = { 0 };
+	ConnStrings *dsn = &(listDBoptions.connStrings);
 
-	(void) parse_and_scrub_connection_string(listDBoptions.source_pguri,
-											 scrubbedSourceURI);
+	log_info("Listing indexes in \"%s\"", dsn->safeSourcePGURI.pguri);
 
-	log_info("Listing indexes in \"%s\"", scrubbedSourceURI);
-
-	if (!pgsql_init(&pgsql, listDBoptions.source_pguri, PGSQL_CONN_SOURCE))
+	if (!pgsql_init(&pgsql, dsn->source_pguri, PGSQL_CONN_SOURCE))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_SOURCE);
@@ -903,8 +1428,8 @@ cli_list_indexes(int argc, char **argv)
 						index->indexNamespace,
 						index->indexRelname,
 						index->constraintName,
-						index->constraintDef,
-						index->indexDef);
+						NULL_AS_EMPTY_STRING(index->constraintDef),
+						NULL_AS_EMPTY_STRING(index->indexDef));
 			}
 			else
 			{
@@ -918,7 +1443,7 @@ cli_list_indexes(int argc, char **argv)
 						index->indexNamespace,
 						"",
 						index->constraintName,
-						index->constraintDef,
+						NULL_AS_EMPTY_STRING(index->constraintDef),
 						"");
 			}
 		}
@@ -930,8 +1455,8 @@ cli_list_indexes(int argc, char **argv)
 					index->indexNamespace,
 					index->indexRelname,
 					index->constraintName,
-					index->constraintDef,
-					index->indexDef);
+					NULL_AS_EMPTY_STRING(index->constraintDef),
+					NULL_AS_EMPTY_STRING(index->indexDef));
 		}
 	}
 
@@ -972,7 +1497,7 @@ cli_list_depends(int argc, char **argv)
 
 			if (filters.type == SOURCE_FILTER_TYPE_NONE)
 			{
-				log_error("BUG: can't list skipped sequences "
+				log_error("BUG: can't list skipped dependencies "
 						  " from filtering type %d",
 						  filters.type);
 				exit(EXIT_CODE_INTERNAL_ERROR);
@@ -980,7 +1505,9 @@ cli_list_depends(int argc, char **argv)
 		}
 	}
 
-	if (!pgsql_init(&pgsql, listDBoptions.source_pguri, PGSQL_CONN_SOURCE))
+	ConnStrings *dsn = &(listDBoptions.connStrings);
+
+	if (!pgsql_init(&pgsql, dsn->source_pguri, PGSQL_CONN_SOURCE))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_SOURCE);
@@ -1034,36 +1561,29 @@ cli_list_schema(int argc, char **argv)
 	 * Assume --resume so that we can run the command alongside the main
 	 * process being active.
 	 */
-	bool auxilliary = true;
+	bool createWorkDir = true;
 
 	if (!copydb_init_workdir(&copySpecs,
 							 NULL,
+							 false, /* service */
+							 NULL,  /* serviceName */
 							 false, /* restart */
 							 true, /* resume */
-							 auxilliary))
+							 createWorkDir))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
-	RestoreOptions restoreOptions = { 0 };
+	CopyDBOptions options = { 0 };
 
-	if (!copydb_init_specs(&copySpecs,
-						   listDBoptions.source_pguri,
-						   "",  /* target_pguri */
-						   1,   /* tableJobs */
-						   1,   /* indexJobs */
-						   listDBoptions.splitTablesLargerThan,
-						   listDBoptions.splitTablesLargerThanPretty,
-						   DATA_SECTION_ALL,
-						   "",  /* snapshot */
-						   restoreOptions,
-						   false, /* roles */
-						   false, /* skipLargeObjects */
-						   false, /* skipExtensions */
-						   false, /* restart */
-						   true,  /* resume */
-						   false)) /* consistent */
+	if (!copydb_init_specs_from_listdboptions(&options, &listDBoptions))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_BAD_ARGS);
+	}
+
+	if (!copydb_init_specs(&copySpecs, &options, DATA_SECTION_ALL))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_INTERNAL_ERROR);
@@ -1080,12 +1600,20 @@ cli_list_schema(int argc, char **argv)
 		}
 	}
 
-	char scrubbedSourceURI[MAXCONNINFO] = { 0 };
+	/*
+	 * First, we need to open a snapshot that we're going to re-use in all our
+	 * connections to the source database. When the --snapshot option has been
+	 * used, instead of exporting a new snapshot, we can just re-use it.
+	 */
+	if (!copydb_prepare_snapshot(&copySpecs))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_INTERNAL_ERROR);
+	}
 
-	(void) parse_and_scrub_connection_string(copySpecs.source_pguri,
-											 scrubbedSourceURI);
+	ConnStrings *dsn = &(listDBoptions.connStrings);
 
-	log_info("Fetching schema from \"%s\"", scrubbedSourceURI);
+	log_info("Fetching schema from \"%s\"", dsn->safeSourcePGURI.pguri);
 	log_info("Dumping schema into JSON file \"%s\"",
 			 copySpecs.cfPaths.schemafile);
 
@@ -1128,45 +1656,80 @@ cli_list_progress(int argc, char **argv)
 
 	(void) find_pg_commands(&(copySpecs.pgPaths));
 
+	char *dir =
+		IS_EMPTY_STRING_BUFFER(listDBoptions.dir)
+		? NULL
+		: listDBoptions.dir;
+
 	/*
 	 * Assume --resume so that we can run the command alongside the main
 	 * process being active.
 	 */
-	bool auxilliary = true;
+	bool createWorkDir = false;
 
 	if (!copydb_init_workdir(&copySpecs,
-							 NULL,
+							 dir,
+							 false, /* service */
+							 NULL,  /* serviceName */
 							 false, /* restart */
 							 true, /* resume */
-							 auxilliary))
+							 createWorkDir))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
-	RestoreOptions restoreOptions = { 0 };
+	CopyDBOptions options = { 0 };
 
-	if (!copydb_init_specs(&copySpecs,
-						   listDBoptions.source_pguri,
-						   "",  /* target_pguri */
-						   1,   /* tableJobs */
-						   1,   /* indexJobs */
-						   listDBoptions.splitTablesLargerThan,
-						   listDBoptions.splitTablesLargerThanPretty,
-						   DATA_SECTION_ALL,
-						   "",  /* snapshot */
-						   restoreOptions,
-						   false, /* roles */
-						   false, /* skipLargeObjects */
-						   false, /* skipExtensions */
-						   false, /* restart */
-						   true,  /* resume */
-						   false)) /* consistent */
+	if (!copydb_init_specs_from_listdboptions(&options, &listDBoptions))
+	{
+		/* errors have already been logged */
+		exit(EXIT_CODE_BAD_ARGS);
+	}
+
+	if (!copydb_init_specs(&copySpecs, &options, DATA_SECTION_ALL))
 	{
 		/* errors have already been logged */
 		exit(EXIT_CODE_INTERNAL_ERROR);
 	}
 
+	if (listDBoptions.summary)
+	{
+		if (outputJSON)
+		{
+			const char *filename = copySpecs.cfPaths.summaryfile;
+
+			if (!file_exists(filename))
+			{
+				log_fatal("Summary JSON file \"%s\" does not exists", filename);
+				exit(EXIT_CODE_INTERNAL_ERROR);
+			}
+
+			char *fileContents = NULL;
+			long fileSize = 0L;
+
+			if (!read_file(filename, &fileContents, &fileSize))
+			{
+				/* errors have already been logged */
+				exit(EXIT_CODE_INTERNAL_ERROR);
+			}
+
+			fformat(stdout, "%s\n", fileContents);
+
+			exit(EXIT_CODE_QUIT);
+		}
+		else
+		{
+			/*
+			 * TODO: parse the JSON summary file, prepare our internal data
+			 * structure with the information found, including pretty printed
+			 * strings for durations etc, and then call print_summary().
+			 */
+			log_fatal("Failed to display summary, please use --json");
+			exit(EXIT_CODE_BAD_ARGS);
+		}
+	}
+
 	if (!copydb_parse_schema_json_file(&copySpecs))
 	{
 		/* errors have already been logged */
@@ -1225,3 +1788,22 @@ cli_list_progress(int argc, char **argv)
 				progress.indexDoneCount);
 	}
 }
+
+
+/*
+ * copydb_init_specs_from_listdboptions initializes a CopyDBOptions structure
+ * from a listDBoptions structure.
+ */
+static bool
+copydb_init_specs_from_listdboptions(CopyDBOptions *options,
+									 ListDBOptions *listDBoptions)
+{
+	strlcpy(options->dir, listDBoptions->dir, MAXPGPATH);
+	options->connStrings = listDBoptions->connStrings;
+	options->splitTablesLargerThan = listDBoptions->splitTablesLargerThan;
+
+	/* pretend like --resume was used in every `pgcopydb list ...` commands */
+	options->resume = true;
+
+	return true;
+}
