@@ -17,16 +17,23 @@
 
 typedef struct ListDBOptions
 {
-	char source_pguri[MAXCONNINFO];
+	char dir[MAXPGPATH];
+
+	ConnStrings connStrings;
+
 	char schema_name[NAMEDATALEN];
 	char table_name[NAMEDATALEN];
 	char filterFileName[MAXPGPATH];
 
 	bool listSkipped;
 	bool noPKey;
+	bool cache;
+	bool dropCache;
+	bool summary;
+	bool availableVersions;
+	bool requirements;
 
-	uint64_t splitTablesLargerThan;
-	char splitTablesLargerThanPretty[NAMEDATALEN];
+	SplitTableLargerThan splitTablesLargerThan;
 } ListDBOptions;
 
 
