@@ -9,7 +9,7 @@
 #include "git-version.h"
 
 /* additional version information for printing version on CLI */
-#define PGCOPYDB_VERSION "0.10"
+#define PGCOPYDB_VERSION "0.13"
 
 #ifdef GIT_VERSION
 #define VERSION_STRING GIT_VERSION
@@ -31,6 +31,14 @@
 #define PGCOPYDB_SPLIT_TABLES_LARGER_THAN "PGCOPYDB_SPLIT_TABLES_LARGER_THAN"
 #define PGCOPYDB_DROP_IF_EXISTS "PGCOPYDB_DROP_IF_EXISTS"
 #define PGCOPYDB_SNAPSHOT "PGCOPYDB_SNAPSHOT"
+#define PGCOPYDB_OUTPUT_PLUGIN "PGCOPYDB_OUTPUT_PLUGIN"
+#define PGCOPYDB_LOG_TIME_FORMAT "PGCOPYDB_LOG_TIME_FORMAT"
+#define PGCOPYDB_LOG_JSON "PGCOPYDB_LOG_JSON"
+#define PGCOPYDB_LOG_JSON_FILE "PGCOPYDB_LOG_JSON_FILE"
+#define PGCOPYDB_LOG_FILENAME "PGCOPYDB_LOG_FILENAME"
+#define PGCOPYDB_FAIL_FAST "PGCOPYDB_FAIL_FAST"
+#define PGCOPYDB_SKIP_VACUUM "PGCOPYDB_SKIP_VACUUM"
+
 #define PGCOPYDB_PGAPPNAME "pgcopydb"
 
 /* default values for the command line options */
@@ -53,10 +61,11 @@
 
 /* default replication slot and origin for logical replication */
 #define REPLICATION_ORIGIN "pgcopydb"
-#define REPLICATION_PLUGIN "wal2json"
+#define REPLICATION_PLUGIN "test_decoding"
 #define REPLICATION_SLOT_NAME "pgcopydb"
 
-#define CATCHINGUP_SLEEP_MS 10 * 1000 /* 10s */
+#define CATCHINGUP_SLEEP_MS 1 * 1000 /* 1s */
+#define STREAM_EMPTY_TX_TIMEOUT 10   /* seconds */
 
 /* internal default for allocating strings  */
 #define BUFSIZE 1024
