@@ -47,6 +47,13 @@ copydb_dump_source_schema(CopyDataSpec *specs,
 						  const char *snapshot,
 						  PostgresDumpSection section)
 {
+	SourceExtensionArray *extensionArray = NULL;
+
+	if (specs->skipExtensions)
+	{
+		extensionArray = &(specs->catalog.extensionArray);
+	}
+
 	if (section == PG_DUMP_SECTION_SCHEMA ||
 		section == PG_DUMP_SECTION_PRE_DATA ||
 		section == PG_DUMP_SECTION_ALL)
@@ -58,9 +65,11 @@ copydb_dump_source_schema(CopyDataSpec *specs,
 					 specs->cfPaths.done.preDataDump);
 		}
 		else if (!pg_dump_db(&(specs->pgPaths),
-							 specs->source_pguri,
+							 &(specs->connStrings),
 							 snapshot,
 							 "pre-data",
+							 &(specs->filters),
+							 extensionArray,
 							 specs->dumpPaths.preFilename))
 		{
 			/* errors have already been logged */
@@ -70,7 +79,7 @@ copydb_dump_source_schema(CopyDataSpec *specs,
 		/* now write the doneFile to keep track */
 		if (!write_file("", 0, specs->cfPaths.done.preDataDump))
 		{
-			log_error("Failed to write the tracking file \%s\"",
+			log_error("Failed to write the tracking file \"%s\"",
 					  specs->cfPaths.done.preDataDump);
 			return false;
 		}
@@ -87,9 +96,11 @@ copydb_dump_source_schema(CopyDataSpec *specs,
 					 specs->cfPaths.done.postDataDump);
 		}
 		else if (!pg_dump_db(&(specs->pgPaths),
-							 specs->source_pguri,
+							 &(specs->connStrings),
 							 snapshot,
 							 "post-data",
+							 &(specs->filters),
+							 extensionArray,
 							 specs->dumpPaths.postFilename))
 		{
 			/* errors have already been logged */
@@ -99,7 +110,7 @@ copydb_dump_source_schema(CopyDataSpec *specs,
 		/* now write the doneFile to keep track */
 		if (!write_file("", 0, specs->cfPaths.done.postDataDump))
 		{
-			log_error("Failed to write the tracking file \%s\"",
+			log_error("Failed to write the tracking file \"%s\"",
 					  specs->cfPaths.done.postDataDump);
 			return false;
 		}
@@ -154,7 +165,8 @@ copydb_target_prepare_schema(CopyDataSpec *specs)
 	}
 
 	if (!pg_restore_db(&(specs->pgPaths),
-					   specs->target_pguri,
+					   &(specs->connStrings),
+					   &(specs->filters),
 					   specs->dumpPaths.preFilename,
 					   specs->dumpPaths.preListFilename,
 					   specs->restoreOptions))
@@ -166,7 +178,7 @@ copydb_target_prepare_schema(CopyDataSpec *specs)
 	/* now write the doneFile to keep track */
 	if (!write_file("", 0, specs->cfPaths.done.preDataRestore))
 	{
-		log_error("Failed to write the tracking file \%s\"",
+		log_error("Failed to write the tracking file \"%s\"",
 				  specs->cfPaths.done.preDataRestore);
 		return false;
 	}
@@ -183,20 +195,26 @@ copydb_target_prepare_schema(CopyDataSpec *specs)
 bool
 copydb_target_drop_tables(CopyDataSpec *specs)
 {
-	PQExpBuffer query = createPQExpBuffer();
-
 	log_info("Drop tables on the target database, per --drop-if-exists");
 
-	appendPQExpBuffer(query, "DROP TABLE IF EXISTS");
+	SourceTableArray *tableArray = &(specs->catalog.sourceTableArray);
 
-	SourceTableArray *tableArray = &(specs->sourceTableArray);
+	if (tableArray->count == 0)
+	{
+		log_info("No tables to migrate, skipping drop tables on the target database");
+		return true;
+	}
+
+	PQExpBuffer query = createPQExpBuffer();
+
+	appendPQExpBuffer(query, "DROP TABLE IF EXISTS");
 
 	for (int tableIndex = 0; tableIndex < tableArray->count; tableIndex++)
 	{
 		SourceTable *source = &(tableArray->array[tableIndex]);
 
 		appendPQExpBuffer(query, "%s \"%s\".\"%s\"",
-						  tableIndex == 0 ? " " : ", ",
+						  tableIndex == 0 ? " " : ",",
 						  source->nspname,
 						  source->relname);
 	}
@@ -213,7 +231,7 @@ copydb_target_drop_tables(CopyDataSpec *specs)
 
 	PGSQL dst = { 0 };
 
-	if (!pgsql_init(&dst, specs->target_pguri, PGSQL_CONN_TARGET))
+	if (!pgsql_init(&dst, specs->connStrings.target_pguri, PGSQL_CONN_TARGET))
 	{
 		/* errors have already been logged */
 		destroyPQExpBuffer(query);
@@ -261,7 +279,8 @@ copydb_target_finalize_schema(CopyDataSpec *specs)
 	}
 
 	if (!pg_restore_db(&(specs->pgPaths),
-					   specs->target_pguri,
+					   &(specs->connStrings),
+					   &(specs->filters),
 					   specs->dumpPaths.postFilename,
 					   specs->dumpPaths.postListFilename,
 					   specs->restoreOptions))
@@ -273,7 +292,7 @@ copydb_target_finalize_schema(CopyDataSpec *specs)
 	/* now write the doneFile to keep track */
 	if (!write_file("", 0, specs->cfPaths.done.postDataRestore))
 	{
-		log_error("Failed to write the tracking file \%s\"",
+		log_error("Failed to write the tracking file \"%s\"",
 				  specs->cfPaths.done.postDataRestore);
 		return false;
 	}
@@ -348,20 +367,59 @@ copydb_write_restore_list(CopyDataSpec *specs, PostgresDumpSection section)
 	if (listContents == NULL)
 	{
 		log_error(ALLOCATION_FAILED_ERROR);
-		free(listContents);
+		destroyPQExpBuffer(listContents);
 		return false;
 	}
 
 	/* for each object in the list, comment when we already processed it */
 	for (int i = 0; i < contents.count; i++)
 	{
-		uint32_t oid = contents.array[i].objectOid;
-		char *name = contents.array[i].restoreListName;
-		char *prefix = "";
+		ArchiveContentItem *item = &(contents.array[i]);
+		uint32_t oid = item->objectOid;
+		uint32_t catOid = item->catalogOid;
+		char *name = item->restoreListName;
 
-		if (copydb_objectid_has_been_processed_already(specs, oid))
+		bool skip = false;
+
+		/*
+		 * Skip COMMENT ON EXTENSION when either of the option
+		 * --skip-extensions or --skip-ext-comment has been used.
+		 */
+		if ((specs->skipExtensions ||
+			 specs->skipCommentOnExtension) &&
+			item->isCompositeTag &&
+			item->tagKind == ARCHIVE_TAG_KIND_COMMENT &&
+			item->tagType == ARCHIVE_TAG_TYPE_EXTENSION)
 		{
-			prefix = ";";
+			skip = true;
+			log_notice("Skipping COMMENT ON EXTENSION \"%s\"", name);
+		}
+
+		if (!skip && catOid == PG_NAMESPACE_OID)
+		{
+			bool exists = false;
+
+			if (!copydb_schema_already_exists(specs, name, &exists))
+			{
+				log_error("Failed to check if restore name \"%s\" "
+						  "already exists",
+						  name);
+				return false;
+			}
+
+			if (exists)
+			{
+				skip = true;
+
+				log_notice("Skipping already existing schema %u: %s",
+						   contents.array[i].objectOid,
+						   contents.array[i].restoreListName);
+			}
+		}
+
+		if (!skip && copydb_objectid_has_been_processed_already(specs, oid))
+		{
+			skip = true;
 
 			log_notice("Skipping already processed dumpId %d: %s %u %s",
 					   contents.array[i].dumpId,
@@ -369,19 +427,36 @@ copydb_write_restore_list(CopyDataSpec *specs, PostgresDumpSection section)
 					   contents.array[i].objectOid,
 					   contents.array[i].restoreListName);
 		}
-		else if (copydb_objectid_is_filtered_out(specs, oid, name))
-		{
-			prefix = ";";
 
-			log_notice("Skipping filtered-out dumpId %d: %s %u %s",
+		/*
+		 * For SEQUENCE catalog entries, we want to limit the scope of the hash
+		 * table search to the OID, and bypass searching by restore name. We
+		 * only use the restore name for the SEQUENCE OWNED BY statements.
+		 *
+		 * This also allows complex filtering of sequences that are owned by
+		 * table a and used as a default value in table b, where table a has
+		 * been filtered-out from pgcopydb scope of operations, but not table
+		 * b.
+		 */
+		if (streq(item->desc, "SEQUENCE"))
+		{
+			name = NULL;
+		}
+
+		if (!skip && copydb_objectid_is_filtered_out(specs, oid, name))
+		{
+			skip = true;
+
+			log_notice("Skipping filtered-out dumpId %d: %s %u %u %s",
 					   contents.array[i].dumpId,
 					   contents.array[i].desc,
+					   contents.array[i].catalogOid,
 					   contents.array[i].objectOid,
 					   contents.array[i].restoreListName);
 		}
 
 		appendPQExpBuffer(listContents, "%s%d; %u %u %s %s\n",
-						  prefix,
+						  skip ? ";" : "",
 						  contents.array[i].dumpId,
 						  contents.array[i].catalogOid,
 						  contents.array[i].objectOid,
