@@ -29,13 +29,20 @@
  * OID values from PostgreSQL src/include/catalog/pg_type.h
  */
 #define BOOLOID 16
+#define BYTEAOID 17
 #define NAMEOID 19
 #define INT4OID 23
 #define INT8OID 20
 #define TEXTOID 25
+#define OIDOID 26
 #define LSNOID 3220
 #define TIMESTAMPTZOID 1184
 
+/*
+ * Catalog OID values from PostgreSQL src/include/catalog/pg_namespace.h
+ */
+#define PG_NAMESPACE_OID 2615
+
 /*
  * Error codes that we use internally.
  */
