@@ -17,9 +17,11 @@
 #include "defaults.h"
 #include "env_utils.h"
 #include "file_utils.h"
+#include "filtering.h"
 #include "log.h"
 #include "parsing_utils.h"
 #include "pgcmd.h"
+#include "schema.h"
 #include "signals.h"
 #include "string_utils.h"
 
@@ -336,38 +338,32 @@ set_psql_from_pg_config(PostgresPaths *pgPaths)
  */
 bool
 pg_dump_db(PostgresPaths *pgPaths,
-		   const char *pguri,
+		   ConnStrings *connStrings,
 		   const char *snapshot,
 		   const char *section,
+		   SourceFilters *filters,
+		   SourceExtensionArray *extensionArray,
 		   const char *filename)
 {
-	char *args[16];
+	char *args[PG_CMD_MAX_ARG];
 	int argsIndex = 0;
 
 	char command[BUFSIZE] = { 0 };
 
-	SafeURI safeURI = { 0 };
+	char *PGPASSWORD = NULL;
 	bool pgpassword_found_in_env = env_exists("PGPASSWORD");
-	char PGPASSWORD[MAXCONNINFO] = { 0 };
-
-	if (!extract_connection_string_password(pguri, &safeURI))
-	{
-		/* errors have already been logged */
-		return false;
-	}
 
 	setenv("PGCONNECT_TIMEOUT", POSTGRES_CONNECT_TIMEOUT, 1);
 
 	/* override PGPASSWORD environment variable if the pguri contains one */
-	if (!IS_EMPTY_STRING_BUFFER(safeURI.password))
+	if (connStrings->safeSourcePGURI.password != NULL)
 	{
-		if (pgpassword_found_in_env &&
-			!get_env_copy("PGPASSWORD", PGPASSWORD, MAXCONNINFO))
+		if (pgpassword_found_in_env && !get_env_dup("PGPASSWORD", &PGPASSWORD))
 		{
 			/* errors have already been logged */
 			return false;
 		}
-		setenv("PGPASSWORD", safeURI.password, 1);
+		setenv("PGPASSWORD", connStrings->safeSourcePGURI.password, 1);
 	}
 
 	args[argsIndex++] = (char *) pgPaths->pg_dump;
@@ -381,9 +377,73 @@ pg_dump_db(PostgresPaths *pgPaths,
 
 	args[argsIndex++] = "--section";
 	args[argsIndex++] = (char *) section;
+
+	/* apply [include-only-schema] filtering */
+	for (int i = 0; i < filters->includeOnlySchemaList.count; i++)
+	{
+		char *nspname = filters->includeOnlySchemaList.array[i].nspname;
+
+		/* check that we still have room for --includeOnly-schema args */
+		if (PG_CMD_MAX_ARG < (argsIndex + 2))
+		{
+			log_error("Failed to call pg_dump, too many include-only-schema entries: "
+					  "argsIndex %d > %d",
+					  argsIndex + 2, PG_CMD_MAX_ARG);
+			return false;
+		}
+
+		args[argsIndex++] = "--schema";
+		args[argsIndex++] = nspname;
+	}
+
+	/* apply [exclude-schema] filtering */
+	for (int i = 0; i < filters->excludeSchemaList.count; i++)
+	{
+		char *nspname = filters->excludeSchemaList.array[i].nspname;
+
+		/* check that we still have room for --exclude-schema args */
+		if (PG_CMD_MAX_ARG < (argsIndex + 2))
+		{
+			log_error("Failed to call pg_dump, too many exclude-schema entries: "
+					  "argsIndex %d > %d",
+					  argsIndex + 2,
+					  PG_CMD_MAX_ARG);
+			return false;
+		}
+
+		args[argsIndex++] = "--exclude-schema";
+		args[argsIndex++] = nspname;
+	}
+
+	/* now --exclude-schema for extension's own schemas */
+	if (extensionArray != NULL)
+	{
+		for (int i = 0; i < extensionArray->count; i++)
+		{
+			char *nspname = extensionArray->array[i].extnamespace;
+
+			if (!streq(nspname, "public") && !streq(nspname, "pg_catalog"))
+			{
+				/* check that we still have room for --exclude-schema args */
+				if (PG_CMD_MAX_ARG < (argsIndex + 2))
+				{
+					log_error("Failed to call pg_dump, "
+							  "too many schema are excluded: "
+							  "argsIndex %d > %d",
+							  argsIndex + 2,
+							  PG_CMD_MAX_ARG);
+					return false;
+				}
+
+				args[argsIndex++] = "--exclude-schema";
+				args[argsIndex++] = nspname;
+			}
+		}
+	}
+
 	args[argsIndex++] = "--file";
 	args[argsIndex++] = (char *) filename;
-	args[argsIndex++] = (char *) safeURI.pguri;
+	args[argsIndex++] = (char *) connStrings->safeSourcePGURI.pguri;
 
 	args[argsIndex] = NULL;
 
@@ -411,7 +471,8 @@ pg_dump_db(PostgresPaths *pgPaths,
 	(void) execute_subprogram(&program);
 
 	/* make sure to reset the environment PGPASSWORD if we edited it */
-	if (pgpassword_found_in_env && !IS_EMPTY_STRING_BUFFER(safeURI.password))
+	if (pgpassword_found_in_env &&
+		connStrings->safeSourcePGURI.password != NULL)
 	{
 		setenv("PGPASSWORD", PGPASSWORD, 1);
 	}
@@ -434,36 +495,29 @@ pg_dump_db(PostgresPaths *pgPaths,
  */
 bool
 pg_dumpall_roles(PostgresPaths *pgPaths,
-				 const char *pguri,
-				 const char *filename)
+				 ConnStrings *connStrings,
+				 const char *filename,
+				 bool noRolesPasswords)
 {
 	char *args[16];
 	int argsIndex = 0;
 
 	char command[BUFSIZE] = { 0 };
 
-	SafeURI safeURI = { 0 };
+	char *PGPASSWORD = NULL;
 	bool pgpassword_found_in_env = env_exists("PGPASSWORD");
-	char PGPASSWORD[MAXCONNINFO] = { 0 };
-
-	if (!extract_connection_string_password(pguri, &safeURI))
-	{
-		/* errors have already been logged */
-		return false;
-	}
 
 	setenv("PGCONNECT_TIMEOUT", POSTGRES_CONNECT_TIMEOUT, 1);
 
 	/* override PGPASSWORD environment variable if the pguri contains one */
-	if (!IS_EMPTY_STRING_BUFFER(safeURI.password))
+	if (connStrings->safeSourcePGURI.password != NULL)
 	{
-		if (pgpassword_found_in_env &&
-			!get_env_copy("PGPASSWORD", PGPASSWORD, MAXCONNINFO))
+		if (pgpassword_found_in_env && !get_env_dup("PGPASSWORD", &PGPASSWORD))
 		{
 			/* errors have already been logged */
 			return false;
 		}
-		setenv("PGPASSWORD", safeURI.password, 1);
+		setenv("PGPASSWORD", connStrings->safeSourcePGURI.password, 1);
 	}
 
 	args[argsIndex++] = (char *) pgPaths->pg_dumpall;
@@ -473,7 +527,12 @@ pg_dumpall_roles(PostgresPaths *pgPaths,
 	args[argsIndex++] = (char *) filename;
 
 	args[argsIndex++] = "--dbname";
-	args[argsIndex++] = (char *) safeURI.pguri;
+	args[argsIndex++] = (char *) connStrings->safeSourcePGURI.pguri;
+
+	if (noRolesPasswords)
+	{
+		args[argsIndex++] = "--no-role-passwords";
+	}
 
 	args[argsIndex] = NULL;
 
@@ -501,7 +560,8 @@ pg_dumpall_roles(PostgresPaths *pgPaths,
 	(void) execute_subprogram(&program);
 
 	/* make sure to reset the environment PGPASSWORD if we edited it */
-	if (pgpassword_found_in_env && !IS_EMPTY_STRING_BUFFER(safeURI.password))
+	if (pgpassword_found_in_env &&
+		connStrings->safeSourcePGURI.password != NULL)
 	{
 		setenv("PGPASSWORD", PGPASSWORD, 1);
 	}
@@ -689,17 +749,17 @@ pg_restore_roles(PostgresPaths *pgPaths,
  */
 bool
 pg_copy_roles(PostgresPaths *pgPaths,
-			  const char *source_pguri,
-			  const char *target_pguri,
-			  const char *filename)
+			  ConnStrings *connStrings,
+			  const char *filename,
+			  bool noRolesPasswords)
 {
-	if (!pg_dumpall_roles(pgPaths, source_pguri, filename))
+	if (!pg_dumpall_roles(pgPaths, connStrings, filename, noRolesPasswords))
 	{
 		/* errors have already been logged */
 		return false;
 	}
 
-	if (!pg_restore_roles(pgPaths, target_pguri, filename))
+	if (!pg_restore_roles(pgPaths, connStrings->target_pguri, filename))
 	{
 		/* errors have already been logged */
 		return false;
@@ -715,43 +775,36 @@ pg_copy_roles(PostgresPaths *pgPaths,
  */
 bool
 pg_restore_db(PostgresPaths *pgPaths,
-			  const char *pguri,
+			  ConnStrings *connStrings,
+			  SourceFilters *filters,
 			  const char *dumpFilename,
 			  const char *listFilename,
 			  RestoreOptions options)
 {
-	char *args[16];
+	char *args[PG_CMD_MAX_ARG];
 	int argsIndex = 0;
 
 	char command[BUFSIZE] = { 0 };
 
-	SafeURI safeURI = { 0 };
+	char *PGPASSWORD = NULL;
 	bool pgpassword_found_in_env = env_exists("PGPASSWORD");
-	char PGPASSWORD[MAXCONNINFO] = { 0 };
-
-	if (!extract_connection_string_password(pguri, &safeURI))
-	{
-		/* errors have already been logged */
-		return false;
-	}
 
 	setenv("PGCONNECT_TIMEOUT", POSTGRES_CONNECT_TIMEOUT, 1);
 
 	/* override PGPASSWORD environment variable if the pguri contains one */
-	if (!IS_EMPTY_STRING_BUFFER(safeURI.password))
+	if (connStrings->safeSourcePGURI.password != NULL)
 	{
-		if (pgpassword_found_in_env &&
-			!get_env_copy("PGPASSWORD", PGPASSWORD, MAXCONNINFO))
+		if (pgpassword_found_in_env && !get_env_dup("PGPASSWORD", &PGPASSWORD))
 		{
 			/* errors have already been logged */
 			return false;
 		}
-		setenv("PGPASSWORD", safeURI.password, 1);
+		setenv("PGPASSWORD", connStrings->safeTargetPGURI.password, 1);
 	}
 
 	args[argsIndex++] = (char *) pgPaths->pg_restore;
 	args[argsIndex++] = "--dbname";
-	args[argsIndex++] = (char *) safeURI.pguri;
+	args[argsIndex++] = (char *) connStrings->safeTargetPGURI.pguri;
 	args[argsIndex++] = "--single-transaction";
 
 	if (options.dropIfExists)
@@ -775,6 +828,43 @@ pg_restore_db(PostgresPaths *pgPaths,
 		args[argsIndex++] = "--no-acl";
 	}
 
+	/* apply [include-only-schema] filtering */
+	for (int i = 0; i < filters->includeOnlySchemaList.count; i++)
+	{
+		char *nspname = filters->includeOnlySchemaList.array[i].nspname;
+
+		/* check that we still have room for --includeOnly-schema args */
+		if (PG_CMD_MAX_ARG < (argsIndex + 2))
+		{
+			log_error("Failed to call pg_dump, too many include-only-schema entries: "
+					  "argsIndex %d > %d",
+					  argsIndex + 2, PG_CMD_MAX_ARG);
+			return false;
+		}
+
+		args[argsIndex++] = "--schema";
+		args[argsIndex++] = nspname;
+	}
+
+	/* apply [exclude-schema] filtering */
+	for (int i = 0; i < filters->excludeSchemaList.count; i++)
+	{
+		char *nspname = filters->excludeSchemaList.array[i].nspname;
+
+		/* check that we still have room for --exclude-schema args */
+		if (PG_CMD_MAX_ARG < (argsIndex + 2))
+		{
+			log_error("Failed to call pg_restore, too many exclude-schema entries: "
+					  "argsIndex %d > %d",
+					  argsIndex + 2,
+					  PG_CMD_MAX_ARG);
+			return false;
+		}
+
+		args[argsIndex++] = "--exclude-schema";
+		args[argsIndex++] = nspname;
+	}
+
 	if (listFilename != NULL)
 	{
 		args[argsIndex++] = "--use-list";
@@ -809,7 +899,8 @@ pg_restore_db(PostgresPaths *pgPaths,
 	(void) execute_subprogram(&program);
 
 	/* make sure to reset the environment PGPASSWORD if we edited it */
-	if (pgpassword_found_in_env && !IS_EMPTY_STRING_BUFFER(safeURI.password))
+	if (pgpassword_found_in_env &&
+		connStrings->safeSourcePGURI.password != NULL)
 	{
 		setenv("PGPASSWORD", PGPASSWORD, 1);
 	}
@@ -953,9 +1044,18 @@ struct StringWithLength pgRestoreDescriptionArray[] = {
 bool
 parse_archive_list(char *list, ArchiveContentArray *contents)
 {
-	/* only parse the first 128 * 1024 lines */
-	char *lines[128 * BUFSIZE];
-	int lineCount = splitLines(list, lines, 128 * BUFSIZE);
+	int lineCount = countLines(list);
+	char **lines = (char **) calloc(lineCount, sizeof(char *));
+	int splitCount = splitLines(list, lines, lineCount);
+
+	if (splitCount != lineCount)
+	{
+		log_error("BUG: parse_archive_list counted %d lines "
+				  "and got %d after split",
+				  lineCount,
+				  splitCount);
+		return false;
+	}
 
 	/* the pg_restore --list preamble is 15 lines long */
 	int objectCount = lineCount - 15;
@@ -1043,6 +1143,8 @@ parse_archive_list(char *list, ArchiveContentArray *contents)
 				if (strcmp(pgRestoreDescriptionArray[i].str, "ACL") == 0 ||
 					strcmp(pgRestoreDescriptionArray[i].str, "COMMENT") == 0)
 				{
+					item->isCompositeTag = true;
+
 					/* ignore errors */
 					if (!parse_archive_acl_or_comment(ptr, item))
 					{
@@ -1107,11 +1209,15 @@ parse_archive_acl_or_comment(char *ptr, ArchiveContentItem *item)
 	{
 		ptr += aclPrefixLen;
 		strlcpy(item->desc, "ACL", sizeof(item->desc));
+
+		item->tagKind = ARCHIVE_TAG_KIND_ACL;
 	}
 	else if (strncmp(ptr, commentPrefix, commentPrefixLen) == 0)
 	{
 		ptr += commentPrefixLen;
 		strlcpy(item->desc, "COMMENT", sizeof(item->desc));
+
+		item->tagKind = ARCHIVE_TAG_KIND_COMMENT;
 	}
 	else
 	{
@@ -1156,6 +1262,8 @@ parse_archive_acl_or_comment(char *ptr, ArchiveContentItem *item)
 		sformat(item->restoreListName, sizeof(item->restoreListName),
 				"- %s",
 				nsp_rol_name);
+
+		item->tagType = ARCHIVE_TAG_TYPE_SCHEMA;
 	}
 	else if (strncmp(ptr, EXTENSION, EXTENSION_LEN) == 0)
 	{
@@ -1172,6 +1280,8 @@ parse_archive_acl_or_comment(char *ptr, ArchiveContentItem *item)
 		sformat(item->restoreListName, sizeof(item->restoreListName),
 				"%s",
 				extname);
+
+		item->tagType = ARCHIVE_TAG_TYPE_EXTENSION;
 	}
 	else
 	{
@@ -1181,6 +1291,9 @@ parse_archive_acl_or_comment(char *ptr, ArchiveContentItem *item)
 		log_debug("Failed to parse %s for %s: not supported yet",
 				  item->desc,
 				  ptr);
+
+		item->tagType = ARCHIVE_TAG_TYPE_OTHER;
+
 		return false;
 	}
 
