@@ -15,8 +15,12 @@
 
 #include "defaults.h"
 #include "file_utils.h"
+#include "filtering.h"
+#include "parsing_utils.h"
 #include "pgsql.h"
+#include "schema.h"
 
+#define PG_CMD_MAX_ARG 128
 #define PG_VERSION_STRING_MAX 12
 
 typedef struct PostgresPaths
@@ -30,6 +34,22 @@ typedef struct PostgresPaths
 } PostgresPaths;
 
 
+typedef enum
+{
+	ARCHIVE_TAG_KIND_UNKNOWN = 0,
+	ARCHIVE_TAG_KIND_ACL,
+	ARCHIVE_TAG_KIND_COMMENT
+} ArchiveCompositeTagKind;
+
+
+typedef enum
+{
+	ARCHIVE_TAG_TYPE_UNKNOWN = 0,
+	ARCHIVE_TAG_TYPE_SCHEMA,
+	ARCHIVE_TAG_TYPE_EXTENSION,
+	ARCHIVE_TAG_TYPE_OTHER
+} ArchiveCompositeTagType;
+
 /*
  * The Postgres pg_restore tool allows listing the contents of an archive. The
  * archive content is formatted the following way:
@@ -54,6 +74,10 @@ typedef struct ArchiveContentItem
 	uint32_t objectOid;
 	char desc[BUFSIZE];
 	char restoreListName[BUFSIZE];
+
+	bool isCompositeTag;
+	ArchiveCompositeTagKind tagKind;
+	ArchiveCompositeTagType tagType;
 } ArchiveContentItem;
 
 
@@ -81,26 +105,30 @@ bool set_psql_from_config_bindir(PostgresPaths *pgPaths, const char *pg_config);
 bool set_psql_from_pg_config(PostgresPaths *pgPaths);
 
 bool pg_dump_db(PostgresPaths *pgPaths,
-				const char *pguri,
+				ConnStrings *connStrings,
 				const char *snapshot,
 				const char *section,
+				SourceFilters *filters,
+				SourceExtensionArray *extensionArray,
 				const char *filename);
 
 bool pg_dumpall_roles(PostgresPaths *pgPaths,
-					  const char *pguri,
-					  const char *filename);
+					  ConnStrings *connStrings,
+					  const char *filename,
+					  bool noRolesPasswords);
 
 bool pg_restore_roles(PostgresPaths *pgPaths,
 					  const char *pguri,
 					  const char *filename);
 
 bool pg_copy_roles(PostgresPaths *pgPaths,
-				   const char *source_pguri,
-				   const char *target_pguri,
-				   const char *filename);
+				   ConnStrings *connStrings,
+				   const char *filename,
+				   bool noRolesPasswords);
 
 bool pg_restore_db(PostgresPaths *pgPaths,
-				   const char *pguri,
+				   ConnStrings *connStrings,
+				   SourceFilters *filters,
 				   const char *dumpFilename,
 				   const char *listFilename,
 				   RestoreOptions options);
