@@ -8,6 +8,7 @@
 
 #include <stdbool.h>
 
+#include "parson.h"
 #include "uthash.h"
 
 #include "filtering.h"
@@ -16,11 +17,33 @@
 /* the pg_restore -l output uses "schema name owner" */
 #define RESTORE_LIST_NAMEDATALEN (3 * NAMEDATALEN + 3)
 
+/*
+ * In the SQL standard we have "catalogs", which are then Postgres databases.
+ * Much the same confusion as with namespace vs schema.
+ */
+typedef struct SourceDatabase
+{
+	uint32_t oid;
+	char datname[NAMEDATALEN];
+	int64_t bytes;
+	char bytesPretty[NAMEDATALEN]; /* pg_size_pretty */
+}
+SourceDatabase;
+
+typedef struct SourceDatabaseArray
+{
+	int count;
+	SourceDatabase *array;
+} SourceDatabaseArray;
+
+
 typedef struct SourceSchema
 {
 	uint32_t oid;
 	char nspname[NAMEDATALEN];
 	char restoreListName[RESTORE_LIST_NAMEDATALEN];
+
+	UT_hash_handle hh;          /* makes this structure hashable */
 } SourceSchema;
 
 typedef struct SourceSchemaArray
@@ -67,6 +90,35 @@ typedef struct SourceExtensionArray
 } SourceExtensionArray;
 
 
+typedef struct ExtensionsVersions
+{
+	char name[NAMEDATALEN];
+	char defaultVersion[BUFSIZE];
+	char installedVersion[BUFSIZE];
+	JSON_Value *json;           /* malloc'ed area */
+} ExtensionsVersions;
+
+typedef struct ExtensionsVersionsArray
+{
+	int count;
+	ExtensionsVersions *array;  /* malloc'ed area */
+} ExtensionsVersionsArray;
+
+typedef struct SourceCollation
+{
+	uint32_t oid;
+	char collname[NAMEDATALEN];
+	char *desc;                 /* malloc'ed area */
+	char restoreListName[RESTORE_LIST_NAMEDATALEN];
+} SourceCollation;
+
+typedef struct SourceCollationArray
+{
+	int count;
+	SourceCollation *array;         /* malloc'ed area */
+} SourceCollationArray;
+
+
 /*
  * SourceTable caches the information we need about all the ordinary tables
  * found in the source database.
@@ -90,12 +142,29 @@ typedef struct SourceTablePartsArray
 } SourceTablePartsArray;
 
 
+typedef struct SourceTableAttribute
+{
+	int attnum;
+	uint32_t atttypid;
+	char attname[NAMEDATALEN];
+	bool attisprimary;
+} SourceTableAttribute;
+
+typedef struct SourceTableAttributeArray
+{
+	int count;
+	SourceTableAttribute *array; /* malloc'ed area */
+} SourceTableAttributeArray;
+
 /* forward declaration */
 struct SourceIndexList;
 
 typedef struct SourceTable
 {
 	uint32_t oid;
+
+	/* "nspname"."relname" : 4 quotes, 1 dot, 1 \0 */
+	char qname[NAMEDATALEN * 2 + 5 + 1];
 	char nspname[NAMEDATALEN];
 	char relname[NAMEDATALEN];
 	int64_t reltuples;
@@ -103,14 +172,20 @@ typedef struct SourceTable
 	char bytesPretty[NAMEDATALEN]; /* pg_size_pretty */
 	bool excludeData;
 
+	uint64_t rowcount;
+	uint64_t checksum;
+
 	char restoreListName[RESTORE_LIST_NAMEDATALEN];
 	char partKey[NAMEDATALEN];
 	SourceTablePartsArray partsArray;
 
+	SourceTableAttributeArray attributes;
+
 	struct SourceIndexList *firstIndex;
 	struct SourceIndexList *lastIndex;
 
 	UT_hash_handle hh;          /* makes this structure hashable */
+	UT_hash_handle hhQName;     /* makes this structure hashable */
 } SourceTable;
 
 
@@ -128,12 +203,20 @@ typedef struct SourceTableArray
 typedef struct SourceSequence
 {
 	uint32_t oid;
+	uint32_t ownedby;           /* pg_class oid of OWNED BY table */
+	uint32_t attrelid;          /* pg_class oid of table using as DEFAULT */
+	uint32_t attroid;           /* pg_attrdef DEFAULT value OID */
+
+	char qname[NAMEDATALEN * 2 + 5 + 1];
 	char nspname[NAMEDATALEN];
 	char relname[NAMEDATALEN];
 	int64_t lastValue;
 	bool isCalled;
 
 	char restoreListName[RESTORE_LIST_NAMEDATALEN];
+
+	UT_hash_handle hh;          /* makes this structure hashable */
+	UT_hash_handle hhQName;     /* makes this structure hashable */
 } SourceSequence;
 
 
@@ -157,11 +240,13 @@ typedef struct SourceIndex
 	char tableRelname[NAMEDATALEN];
 	bool isPrimary;
 	bool isUnique;
-	char indexColumns[BUFSIZE];
-	char indexDef[BUFSIZE];
+	char *indexColumns;         /* malloc'ed area */
+	char *indexDef;             /* malloc'ed area */
 	uint32_t constraintOid;
+	bool condeferrable;
+	bool condeferred;
 	char constraintName[NAMEDATALEN];
-	char constraintDef[BUFSIZE];
+	char *constraintDef;        /* malloc'ed area */
 	char indexRestoreListName[RESTORE_LIST_NAMEDATALEN];
 	char constraintRestoreListName[RESTORE_LIST_NAMEDATALEN];
 
@@ -209,10 +294,57 @@ typedef struct SourceDependArray
 	SourceDepend *array;         /* malloc'ed area */
 } SourceDependArray;
 
+/*
+ * SourceCatalog regroups all the information we fetch from a Postgres
+ * instance.
+ */
+typedef struct SourceCatalog
+{
+	SourceExtensionArray extensionArray;
+	SourceCollationArray collationArray;
+	SourceTableArray sourceTableArray;
+	SourceIndexArray sourceIndexArray;
+	SourceSequenceArray sequenceArray;
+
+	SourceTable *sourceTableHashByOid;
+	SourceTable *sourceTableHashByQName;
+	SourceIndex *sourceIndexHashByOid;
+	SourceSequence *sourceSeqHashByOid;
+	SourceSequence *sourceSeqHashByQname;
+} SourceCatalog;
+
+
+typedef struct TargetCatalog
+{
+	SourceSchemaArray schemaArray;
+	SourceSchema *schemaHashByName;
+} TargetCatalog;
+
+
+bool schema_query_privileges(PGSQL *pgsql,
+							 bool *hasDBCreatePrivilage,
+							 bool *hasDBTempPrivilege);
+
+bool schema_list_databases(PGSQL *pgsql, SourceDatabaseArray *catArray);
+
+bool schema_list_schemas(PGSQL *pgsql, SourceSchemaArray *array);
+
 bool schema_list_ext_schemas(PGSQL *pgsql, SourceSchemaArray *array);
 
 bool schema_list_extensions(PGSQL *pgsql, SourceExtensionArray *extArray);
 
+bool schema_list_ext_versions(PGSQL *pgsql, ExtensionsVersionsArray *array);
+
+bool schema_list_collations(PGSQL *pgsql, SourceCollationArray *array);
+
+bool schema_prepare_pgcopydb_table_size(PGSQL *pgsql,
+										SourceFilters *filters,
+										bool hasDBCreatePrivilege,
+										bool cache,
+										bool dropCache,
+										bool *createdTableSizeTable);
+
+bool schema_drop_pgcopydb_table_size(PGSQL *pgsql);
 
 bool schema_list_ordinary_tables(PGSQL *pgsql,
 								 SourceFilters *filters,
@@ -244,4 +376,6 @@ bool schema_list_pg_depend(PGSQL *pgsql,
 						   SourceFilters *filters,
 						   SourceDependArray *dependArray);
 
+bool schema_checksum_table(PGSQL *pgsql, SourceTable *table);
+
 #endif /* SCHEMA_H */
