@@ -31,21 +31,28 @@ static void prepareLineSeparator(char dashes[], int size);
 bool
 write_table_summary(CopyTableSummary *summary, char *filename)
 {
-	char contents[BUFSIZE] = { 0 };
+	JSON_Value *js = json_value_init_object();
+	JSON_Object *jsObj = json_value_get_object(js);
 
-	sformat(contents, BUFSIZE,
-			"%d\n%u\n%s\n%s\n%lld\n%lld\n%lld\n%s\n",
-			summary->pid,
-			summary->table->oid,
-			summary->table->nspname,
-			summary->table->relname,
-			(long long) summary->startTime,
-			(long long) summary->doneTime,
-			(long long) summary->durationMs,
-			summary->command);
+	json_object_set_number(jsObj, "pid", summary->pid);
+	json_object_dotset_number(jsObj, "table.oid", summary->table->oid);
+	json_object_dotset_string(jsObj, "table.nspname", summary->table->nspname);
+	json_object_dotset_string(jsObj, "table.relname", summary->table->relname);
+	json_object_set_number(jsObj, "start-time-epoch", summary->startTime);
+	json_object_set_number(jsObj, "done-time-epoch", summary->doneTime);
+	json_object_set_number(jsObj, "duration", summary->durationMs);
+	json_object_set_string(jsObj, "command", summary->command);
+
+	char *serialized_string = json_serialize_to_string_pretty(js);
+	size_t len = strlen(serialized_string);
 
 	/* write the summary to the doneFile */
-	if (!write_file(contents, strlen(contents), filename))
+	bool success = write_file(serialized_string, len, filename);
+
+	json_free_serialized_string(serialized_string);
+	json_value_free(js);
+
+	if (!success)
 	{
 		log_error("Failed to write table summary file \"%s\"", filename);
 		return false;
@@ -114,76 +121,39 @@ prepare_table_summary_as_json(CopyTableSummary *summary,
 bool
 read_table_summary(CopyTableSummary *summary, const char *filename)
 {
-	char *fileContents = NULL;
-	long fileSize = 0L;
+	JSON_Value *json = json_parse_file(filename);
 
-	if (!read_file(filename, &fileContents, &fileSize))
+	if (json == NULL)
 	{
-		/* errors have already been logged */
+		log_error("Failed to parse summary file \"%s\"", filename);
 		return false;
 	}
 
-	char *fileLines[BUFSIZE] = { 0 };
-	int lineCount = splitLines(fileContents, fileLines, BUFSIZE);
+	JSON_Object *jsObj = json_value_get_object(json);
 
-	if (lineCount < COPY_TABLE_SUMMARY_LINES)
+	summary->pid = json_object_get_number(jsObj, "pid");
+
+	summary->table->oid = json_object_dotget_number(jsObj, "table.oid");
+
+	char *schema = (char *) json_object_dotget_string(jsObj, "table.nspname");
+	char *name = (char *) json_object_dotget_string(jsObj, "table.relname");
+
+	strlcpy(summary->table->nspname, schema, sizeof(summary->table->nspname));
+	strlcpy(summary->table->relname, name, sizeof(summary->table->relname));
+
+	summary->startTime = json_object_get_number(jsObj, "start-time-epoch");
+	summary->doneTime = json_object_get_number(jsObj, "done-time-epoch");
+	summary->durationMs = json_object_get_number(jsObj, "duration");
+
+	summary->command = strdup(json_object_get_string(jsObj, "command"));
+
+	if (summary->command == NULL)
 	{
-		log_error("Failed to parse summary file \"%s\" which contains only "
-				  "%d lines, at least %d lines are expected",
-				  filename,
-				  lineCount,
-				  COPY_TABLE_SUMMARY_LINES);
-
-		free(fileContents);
-
+		log_error(ALLOCATION_FAILED_ERROR);
+		json_value_free(json);
 		return false;
 	}
 
-	if (!stringToInt(fileLines[0], &(summary->pid)))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	/* better not point to NULL */
-	SourceTable *table = summary->table;
-
-	if (table == NULL)
-	{
-		log_error("BUG: read_table_summary summary->table is NULL");
-		return false;
-	}
-
-	if (!stringToUInt32(fileLines[1], &(table->oid)))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	strlcpy(table->nspname, fileLines[2], sizeof(table->nspname));
-	strlcpy(table->relname, fileLines[3], sizeof(table->relname));
-
-	if (!stringToUInt64(fileLines[4], &(summary->startTime)))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	if (!stringToUInt64(fileLines[5], &(summary->doneTime)))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	if (!stringToUInt64(fileLines[6], &(summary->durationMs)))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	/* last summary line in the file is the SQL command */
-	strlcpy(summary->command, fileLines[7], sizeof(summary->command));
-
 	/* we can't provide instr_time readers */
 	summary->startTimeInstr = (instr_time) {
 		0
@@ -192,6 +162,7 @@ read_table_summary(CopyTableSummary *summary, const char *filename)
 		0
 	};
 
+	json_value_free(json);
 	return true;
 }
 
@@ -260,8 +231,8 @@ create_table_index_file(SourceTable *table, char *filename)
 	{
 		SourceIndex *index = indexListEntry->index;
 
-		appendPQExpBuffer(content, "%d\n", index->indexOid);
-		appendPQExpBuffer(content, "%d\n", index->constraintOid);
+		appendPQExpBuffer(content, "%u\n", index->indexOid);
+		appendPQExpBuffer(content, "%u\n", index->constraintOid);
 	}
 
 	/* memory allocation could have failed while building string */
@@ -314,14 +285,18 @@ read_table_index_file(SourceIndexArray *indexArray, char *filename)
 	 * We expect to have alternate lines with first indexOid and then
 	 * constraintOid (which could be zero). No comments, etc.
 	 */
-	indexArray->count = lineCount / 2;
+	indexArray->count = 0;
 	indexArray->array = (SourceIndex *) calloc(lineCount, sizeof(SourceIndex));
 
-	int pos = 0;
+	if (indexArray->array == NULL)
+	{
+		log_error(ALLOCATION_FAILED_ERROR);
+		return false;
+	}
 
 	for (int i = 0; i < lineCount; i++)
 	{
-		SourceIndex *index = &(indexArray->array[pos]);
+		SourceIndex *index = &(indexArray->array[indexArray->count]);
 
 		uint32_t *target =
 			i % 2 == 0 ? &(index->indexOid) : &(index->constraintOid);
@@ -339,7 +314,7 @@ read_table_index_file(SourceIndexArray *indexArray, char *filename)
 		/* one index entry (indexOid, constraintOid) spans two lines */
 		if (i % 2 == 1)
 		{
-			pos++;
+			++(indexArray->count);
 		}
 	}
 
@@ -355,7 +330,8 @@ read_table_index_file(SourceIndexArray *indexArray, char *filename)
 bool
 write_index_summary(CopyIndexSummary *summary, char *filename, bool constraint)
 {
-	char contents[BUFSIZE] = { 0 };
+	JSON_Value *js = json_value_init_object();
+	JSON_Object *jsObj = json_value_get_object(js);
 
 	uint32_t oid =
 		constraint
@@ -367,19 +343,34 @@ write_index_summary(CopyIndexSummary *summary, char *filename, bool constraint)
 		? summary->index->constraintName
 		: summary->index->indexRelname;
 
-	sformat(contents, BUFSIZE,
-			"%d\n%u\n%s\n%s\n%lld\n%lld\n%lld\n%s\n",
-			summary->pid,
-			oid,
-			summary->index->indexNamespace,
-			name,
-			(long long) summary->startTime,
-			(long long) summary->doneTime,
-			(long long) summary->durationMs,
-			summary->command);
+	json_object_set_number(jsObj, "pid", summary->pid);
+
+	json_object_dotset_number(jsObj, "index.oid", oid);
+	json_object_dotset_string(jsObj, "index.nspname",
+							  summary->index->indexNamespace);
+	json_object_dotset_string(jsObj, "index.relname", name);
+
+	json_object_set_number(jsObj, "start-time-epoch", summary->startTime);
+	json_object_set_number(jsObj, "done-time-epoch", summary->doneTime);
+	json_object_set_number(jsObj, "duration", summary->durationMs);
+	json_object_set_string(jsObj, "command", summary->command);
+
+	char *serialized_string = json_serialize_to_string_pretty(js);
+	size_t len = strlen(serialized_string);
 
 	/* write the summary to the doneFile */
-	return write_file(contents, strlen(contents), filename);
+	bool success = write_file(serialized_string, len, filename);
+
+	json_free_serialized_string(serialized_string);
+	json_value_free(js);
+
+	if (!success)
+	{
+		log_error("Failed to write table summary file \"%s\"", filename);
+		return false;
+	}
+
+	return true;
 }
 
 
@@ -389,76 +380,48 @@ write_index_summary(CopyIndexSummary *summary, char *filename, bool constraint)
 bool
 read_index_summary(CopyIndexSummary *summary, const char *filename)
 {
-	char *fileContents = NULL;
-	long fileSize = 0L;
+	JSON_Value *json = json_parse_file(filename);
 
-	if (!read_file(filename, &fileContents, &fileSize))
+	if (json == NULL)
 	{
-		/* errors have already been logged */
+		log_error("Failed to parse summary file \"%s\"", filename);
 		return false;
 	}
 
-	char *fileLines[BUFSIZE] = { 0 };
-	int lineCount = splitLines(fileContents, fileLines, BUFSIZE);
+	JSON_Object *jsObj = json_value_get_object(json);
 
-	if (lineCount < COPY_TABLE_SUMMARY_LINES)
+	summary->pid = json_object_get_number(jsObj, "pid");
+
+	summary->index->indexOid = json_object_dotget_number(jsObj, "index.oid");
+
+	char *schema = (char *) json_object_dotget_string(jsObj, "index.nspname");
+	char *name = (char *) json_object_dotget_string(jsObj, "index.relname");
+
+	strlcpy(summary->index->indexNamespace,
+			schema,
+			sizeof(summary->index->indexNamespace));
+
+	strlcpy(summary->index->indexRelname,
+			name,
+			sizeof(summary->index->indexRelname));
+
+	summary->startTime = json_object_get_number(jsObj, "start-time-epoch");
+	summary->doneTime = json_object_get_number(jsObj, "done-time-epoch");
+	summary->durationMs = json_object_get_number(jsObj, "duration");
+
+	if (json_object_has_value_of_type(jsObj, "command", JSONString))
 	{
-		log_error("Failed to parse summary file \"%s\" which contains only "
-				  "%d lines, at least %d lines are expected",
-				  filename,
-				  lineCount,
-				  COPY_TABLE_SUMMARY_LINES);
+		const char *command = json_object_get_string(jsObj, "command");
+		summary->command = strdup(command);
 
-		free(fileContents);
-
-		return false;
+		if (summary->command == NULL)
+		{
+			log_error(ALLOCATION_FAILED_ERROR);
+			json_value_free(json);
+			return false;
+		}
 	}
 
-	if (!stringToInt(fileLines[0], &(summary->pid)))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	/* better not point to NULL */
-	SourceIndex *index = summary->index;
-
-	if (index == NULL)
-	{
-		log_error("BUG: read_index_summary summary->index is NULL");
-		return false;
-	}
-
-	if (!stringToUInt32(fileLines[1], &(index->indexOid)))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	strlcpy(index->indexNamespace, fileLines[2], sizeof(index->indexNamespace));
-	strlcpy(index->indexRelname, fileLines[3], sizeof(index->indexRelname));
-
-	if (!stringToUInt64(fileLines[4], &(summary->startTime)))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	if (!stringToUInt64(fileLines[5], &(summary->doneTime)))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	if (!stringToUInt64(fileLines[6], &(summary->durationMs)))
-	{
-		/* errors have already been logged */
-		return false;
-	}
-
-	/* last summary line in the file is the SQL command */
-	strlcpy(summary->command, fileLines[7], sizeof(summary->command));
-
 	/* we can't provide instr_time readers */
 	summary->startTimeInstr = (instr_time) {
 		0
@@ -467,6 +430,7 @@ read_index_summary(CopyIndexSummary *summary, const char *filename)
 		0
 	};
 
+	json_value_free(json);
 	return true;
 }
 
@@ -708,22 +672,24 @@ summary_prepare_toplevel_durations(Summary *summary)
 	uint64_t durationMs;
 
 	/* compute schema dump duration, part of schemaDurationMs */
-	duration = timings->beforeSchemaFetch;
-	INSTR_TIME_SUBTRACT(duration, timings->beforeSchemaDump);
+	duration = timings->beforeSchemaDump;
+	INSTR_TIME_SUBTRACT(duration, timings->beforeSchemaFetch);
 	durationMs = INSTR_TIME_MS(duration);
 
 	IntervalToString(durationMs, timings->dumpSchemaMs, INTSTRING_MAX_DIGITS);
 
 	timings->schemaDurationMs = durationMs;
+	timings->dumpSchemaDurationMs = durationMs;
 
 	/* compute schema fetch duration, part of schemaDurationMs */
 	duration = timings->beforePrepareSchema;
-	INSTR_TIME_SUBTRACT(duration, timings->beforeSchemaFetch);
+	INSTR_TIME_SUBTRACT(duration, timings->beforeSchemaDump);
 	durationMs = INSTR_TIME_MS(duration);
 
 	IntervalToString(durationMs, timings->fetchSchemaMs, INTSTRING_MAX_DIGITS);
 
 	timings->schemaDurationMs += durationMs;
+	timings->fetchSchemaDurationMs = durationMs;
 
 	/* compute prepare schema duration, part of schemaDurationMs */
 	duration = timings->afterPrepareSchema;
@@ -733,6 +699,7 @@ summary_prepare_toplevel_durations(Summary *summary)
 	IntervalToString(durationMs, timings->prepareSchemaMs, INTSTRING_MAX_DIGITS);
 
 	timings->schemaDurationMs += durationMs;
+	timings->prepareSchemaDurationMs = durationMs;
 
 	/* compute data + index duration, between prepare schema and finalize */
 	duration = timings->beforeFinalizeSchema;
@@ -751,6 +718,7 @@ summary_prepare_toplevel_durations(Summary *summary)
 	IntervalToString(durationMs, timings->finalizeSchemaMs, INTSTRING_MAX_DIGITS);
 
 	timings->schemaDurationMs += durationMs;
+	timings->finalizeSchemaDurationMs = durationMs;
 
 	/* compute total duration, wall clock elapsed time */
 	duration = timings->endTime;
@@ -776,7 +744,7 @@ summary_prepare_toplevel_durations(Summary *summary)
  * print_toplevel_summary prints a summary of the top-level timings.
  */
 void
-print_toplevel_summary(Summary *summary, int tableJobs, int indexJobs)
+print_toplevel_summary(Summary *summary)
 {
 	char *d10s = "----------";
 	char *d12s = "------------";
@@ -803,8 +771,8 @@ print_toplevel_summary(Summary *summary, int tableJobs, int indexJobs)
 
 	char concurrency[BUFSIZE] = { 0 };
 	sformat(concurrency, sizeof(concurrency), "%d + %d",
-			tableJobs,
-			tableJobs + indexJobs);
+			summary->tableJobs,
+			summary->tableJobs + summary->indexJobs);
 
 	fformat(stdout, " %50s   %10s  %10s  %12s\n",
 			"COPY, INDEX, CONSTRAINTS, VACUUM (wall clock)", "both",
@@ -814,7 +782,7 @@ print_toplevel_summary(Summary *summary, int tableJobs, int indexJobs)
 	fformat(stdout, " %50s   %10s  %10s  %12d\n",
 			"COPY (cumulative)", "both",
 			summary->timings.totalTableMs,
-			tableJobs);
+			summary->tableJobs);
 
 	fformat(stdout, " %50s   %10s  %10s  %12d\n",
 			"Large Objects (cumulative)", "both",
@@ -824,7 +792,7 @@ print_toplevel_summary(Summary *summary, int tableJobs, int indexJobs)
 	fformat(stdout, " %50s   %10s  %10s  %12d\n",
 			"CREATE INDEX, CONSTRAINTS (cumulative)", "target",
 			summary->timings.totalIndexMs,
-			indexJobs);
+			summary->indexJobs);
 
 	fformat(stdout, " %50s   %10s  %10s  %12d\n", "Finalize Schema", "target",
 			summary->timings.finalizeSchemaMs, 1);
@@ -851,6 +819,11 @@ print_summary_table(SummaryTable *summary)
 {
 	SummaryTableHeaders *headers = &(summary->headers);
 
+	if (summary->count == 0)
+	{
+		return;
+	}
+
 	fformat(stdout, "\n");
 
 	fformat(stdout, "%*s | %*s | %*s | %*s | %*s | %*s\n",
@@ -874,7 +847,7 @@ print_summary_table(SummaryTable *summary)
 		SummaryTableEntry *entry = &(summary->array[i]);
 
 		fformat(stdout, "%*s | %*s | %*s | %*s | %*s | %*s\n",
-				headers->maxOidSize, entry->oid,
+				headers->maxOidSize, entry->oidStr,
 				headers->maxNspnameSize, entry->nspname,
 				headers->maxRelnameSize, entry->relname,
 				headers->maxTableMsSize, entry->tableMs,
@@ -886,6 +859,216 @@ print_summary_table(SummaryTable *summary)
 }
 
 
+/*
+ * print_summary_as_json writes the current summary of operations (with
+ * timings) to given filename, as a structured JSON document.
+ */
+void
+print_summary_as_json(Summary *summary, const char *filename)
+{
+	log_notice("Storing migration summary in JSON file \"%s\"", filename);
+
+	JSON_Value *js = json_value_init_object();
+	JSON_Object *jsobj = json_value_get_object(js);
+
+	json_object_dotset_number(jsobj, "setup.table-jobs", summary->tableJobs);
+	json_object_dotset_number(jsobj, "setup.index-jobs", summary->indexJobs);
+
+	TopLevelTimings *timings = &(summary->timings);
+
+	JSON_Value *jsSteps = json_value_init_array();
+	JSON_Array *jsStepArray = json_value_get_array(jsSteps);
+
+	JSON_Value *jsDumpSchema = json_value_init_object();
+	JSON_Object *jsDSObj = json_value_get_object(jsDumpSchema);
+
+	json_object_set_string(jsDSObj, "label", "dump schema");
+	json_object_set_string(jsDSObj, "conn", "source");
+	json_object_set_number(jsDSObj, "duration",
+						   timings->dumpSchemaDurationMs);
+	json_object_set_number(jsDSObj, "concurrency", 1);
+
+	json_array_append_value(jsStepArray, jsDumpSchema);
+
+	JSON_Value *jsCatalog = json_value_init_object();
+	JSON_Object *jsCatObj = json_value_get_object(jsCatalog);
+
+	json_object_set_string(jsCatObj, "label", "Catalog Queries");
+	json_object_set_string(jsCatObj, "conn", "source");
+	json_object_set_number(jsCatObj, "duration",
+						   timings->fetchSchemaDurationMs);
+	json_object_set_number(jsCatObj, "concurrency", 1);
+
+	json_array_append_value(jsStepArray, jsCatalog);
+
+	JSON_Value *jsPrep = json_value_init_object();
+	JSON_Object *jsPrepObj = json_value_get_object(jsPrep);
+
+	json_object_set_string(jsPrepObj, "label", "Prepare Schema");
+	json_object_set_string(jsPrepObj, "conn", "target");
+	json_object_set_number(jsPrepObj, "duration",
+						   timings->prepareSchemaDurationMs);
+	json_object_set_number(jsPrepObj, "concurrency", 1);
+
+	json_array_append_value(jsStepArray, jsPrep);
+
+	JSON_Value *jsDB = json_value_init_object();
+	JSON_Object *jsDBObj = json_value_get_object(jsDB);
+
+	json_object_set_string(jsDBObj, "label",
+						   "COPY, INDEX, CONSTRAINTS, VACUUM (wall clock)");
+	json_object_set_string(jsDBObj, "conn", "both");
+	json_object_set_number(jsDBObj, "duration",
+						   timings->dataAndIndexesDurationMs);
+	json_object_set_number(jsDBObj, "concurrency",
+						   summary->tableJobs + summary->indexJobs);
+
+	json_array_append_value(jsStepArray, jsDB);
+
+	JSON_Value *jsCopy = json_value_init_object();
+	JSON_Object *jsCopyObj = json_value_get_object(jsCopy);
+
+	json_object_set_string(jsCopyObj, "label", "COPY (cumulative)");
+	json_object_set_string(jsCopyObj, "conn", "both");
+	json_object_set_number(jsCopyObj, "duration", timings->tableDurationMs);
+	json_object_set_number(jsCopyObj, "concurrency", summary->tableJobs);
+
+	json_array_append_value(jsStepArray, jsCopy);
+
+	JSON_Value *jsBlob = json_value_init_object();
+	JSON_Object *jsBlobObj = json_value_get_object(jsBlob);
+
+	json_object_set_string(jsBlobObj, "label", "Large Objects (cumulative)");
+	json_object_set_string(jsBlobObj, "conn", "both");
+	json_object_set_number(jsBlobObj, "duration", timings->blobDurationMs);
+	json_object_set_number(jsBlobObj, "concurrency", 1);
+
+	json_array_append_value(jsStepArray, jsBlob);
+
+	JSON_Value *jsIndex = json_value_init_object();
+	JSON_Object *jsIndexObj = json_value_get_object(jsIndex);
+
+	json_object_set_string(jsIndexObj, "label",
+						   "CREATE INDEX, CONSTRAINTS (cumulative)");
+	json_object_set_string(jsIndexObj, "conn", "target");
+	json_object_set_number(jsIndexObj, "duration", timings->indexDurationMs);
+	json_object_set_number(jsIndexObj, "concurrency", summary->indexJobs);
+
+	json_array_append_value(jsStepArray, jsIndex);
+
+	JSON_Value *jsFin = json_value_init_object();
+	JSON_Object *jsFinObj = json_value_get_object(jsFin);
+
+	json_object_set_string(jsFinObj, "label", "Finalize Schema");
+	json_object_set_string(jsFinObj, "conn", "target");
+	json_object_set_number(jsFinObj, "duration", timings->finalizeSchemaDurationMs);
+	json_object_set_number(jsFinObj, "concurrency", 1);
+
+	json_array_append_value(jsStepArray, jsFin);
+
+	JSON_Value *jsTotal = json_value_init_object();
+	JSON_Object *jsTotalObj = json_value_get_object(jsTotal);
+
+	json_object_set_string(jsTotalObj, "label", "Total Wall Clock Duration");
+	json_object_set_string(jsTotalObj, "conn", "both");
+	json_object_set_number(jsTotalObj, "duration", timings->totalDurationMs);
+
+	json_object_set_number(jsTotalObj,
+						   "concurrency",
+						   summary->tableJobs + summary->indexJobs);
+
+	json_array_append_value(jsStepArray, jsTotal);
+
+	json_object_set_value(jsobj, "steps", jsSteps);
+
+	SummaryTable *summaryTable = &(summary->table);
+
+	JSON_Value *jsTables = json_value_init_array();
+	JSON_Array *jsTableArray = json_value_get_array(jsTables);
+
+	for (int i = 0; i < summaryTable->count; i++)
+	{
+		SummaryTableEntry *entry = &(summaryTable->array[i]);
+
+		JSON_Value *jsTable = json_value_init_object();
+		JSON_Object *jsTableObj = json_value_get_object(jsTable);
+
+		json_object_set_number(jsTableObj, "oid", entry->oid);
+		json_object_set_string(jsTableObj, "schema", entry->nspname);
+		json_object_set_string(jsTableObj, "name", entry->relname);
+
+		json_object_dotset_number(jsTableObj,
+								  "duration", entry->durationTableMs);
+
+		json_object_dotset_number(jsTableObj,
+								  "index.count", entry->indexArray.count);
+		json_object_dotset_number(jsTableObj,
+								  "index.duration", entry->durationIndexMs);
+
+		JSON_Value *jsIndexes = json_value_init_array();
+		JSON_Array *jsIndexArray = json_value_get_array(jsIndexes);
+
+		for (int j = 0; j < entry->indexArray.count; j++)
+		{
+			SummaryIndexEntry *indexEntry = &(entry->indexArray.array[j]);
+
+			JSON_Value *jsIndex = json_value_init_object();
+			JSON_Object *jsIndexObj = json_value_get_object(jsIndex);
+
+			json_object_set_number(jsIndexObj, "oid", indexEntry->oid);
+			json_object_set_string(jsIndexObj, "schema", indexEntry->nspname);
+			json_object_set_string(jsIndexObj, "name", indexEntry->relname);
+			json_object_set_string(jsIndexObj, "sql", indexEntry->sql);
+			json_object_dotset_number(jsIndexObj, "ms", indexEntry->durationMs);
+
+			json_array_append_value(jsIndexArray, jsIndex);
+		}
+
+		/* add the index array to the current table */
+		json_object_set_value(jsTableObj, "indexes", jsIndexes);
+
+		JSON_Value *jsConstraints = json_value_init_array();
+		JSON_Array *jsConstraintArray = json_value_get_array(jsConstraints);
+
+		for (int j = 0; j < entry->constraintArray.count; j++)
+		{
+			SummaryIndexEntry *cEntry = &(entry->constraintArray.array[j]);
+
+			JSON_Value *jsConstraint = json_value_init_object();
+			JSON_Object *jsConstraintObj = json_value_get_object(jsConstraint);
+
+			json_object_set_number(jsConstraintObj, "oid", cEntry->oid);
+			json_object_set_string(jsConstraintObj, "schema", cEntry->nspname);
+			json_object_set_string(jsConstraintObj, "name", cEntry->relname);
+			json_object_set_string(jsConstraintObj, "sql", cEntry->sql);
+			json_object_dotset_number(jsConstraintObj, "ms", cEntry->durationMs);
+
+			json_array_append_value(jsConstraintArray, jsConstraint);
+		}
+
+		/* add the constraint array to the current table */
+		json_object_set_value(jsTableObj, "constraints", jsConstraints);
+
+		/* append the current table to the table array */
+		json_array_append_value(jsTableArray, jsTable);
+	}
+
+	/* add the table array to the main JSON top-level dict */
+	json_object_set_value(jsobj, "tables", jsTables);
+
+	char *serialized_string = json_serialize_to_string_pretty(js);
+	size_t len = strlen(serialized_string);
+
+	if (!write_file(serialized_string, len, filename))
+	{
+		log_error("Failed to write summary JSON file, see above for details");
+	}
+
+	json_free_serialized_string(serialized_string);
+	json_value_free(js);
+}
+
+
 /*
  * prepare_summary_table_headers computes the actual max length of all the
  * columns that we are going to display, and fills in the dashed separators
@@ -910,7 +1093,7 @@ prepare_summary_table_headers(SummaryTable *summary)
 		int len = 0;
 		SummaryTableEntry *entry = &(summary->array[i]);
 
-		len = strlen(entry->oid);
+		len = strlen(entry->oidStr);
 
 		if (headers->maxOidSize < len)
 		{
@@ -997,6 +1180,9 @@ print_summary(Summary *summary, CopyDataSpec *specs)
 {
 	SummaryTable *summaryTable = &(summary->table);
 
+	summary->tableJobs = specs->tableJobs;
+	summary->indexJobs = specs->indexJobs;
+
 	/* first, we have to scan the available data from memory and files */
 	if (!prepare_summary_table(summary, specs))
 	{
@@ -1004,6 +1190,9 @@ print_summary(Summary *summary, CopyDataSpec *specs)
 		return false;
 	}
 
+	/* print the summary.json file */
+	(void) print_summary_as_json(summary, specs->cfPaths.summaryfile);
+
 	/* then we can prepare the headers and print the table */
 	if (specs->section == DATA_SECTION_TABLE_DATA ||
 		specs->section == DATA_SECTION_ALL)
@@ -1014,7 +1203,7 @@ print_summary(Summary *summary, CopyDataSpec *specs)
 
 	/* and then finally prepare the top-level counters and print them */
 	(void) summary_prepare_toplevel_durations(summary);
-	(void) print_toplevel_summary(summary, specs->tableJobs, specs->indexJobs);
+	(void) print_toplevel_summary(summary);
 
 	return true;
 }
@@ -1053,7 +1242,8 @@ prepare_summary_table(Summary *summary, CopyDataSpec *specs)
 		/* prepare some of the information we already have */
 		IntString oidString = intToString(table->oid);
 
-		strlcpy(entry->oid, oidString.strValue, sizeof(entry->oid));
+		entry->oid = table->oid;
+		strlcpy(entry->oidStr, oidString.strValue, sizeof(entry->oidStr));
 		strlcpy(entry->nspname, table->nspname, sizeof(entry->nspname));
 		strlcpy(entry->relname, table->relname, sizeof(entry->relname));
 
@@ -1062,10 +1252,12 @@ prepare_summary_table(Summary *summary, CopyDataSpec *specs)
 
 		if (!read_table_summary(&tableSummary, tableSpecs->tablePaths.doneFile))
 		{
-			/* errors have already been logged */
+			log_error("Failed to read table summary \"%s\"",
+					  tableSpecs->tablePaths.doneFile);
 			return false;
 		}
 
+		entry->durationTableMs = tableSummary.durationMs;
 		timings->tableDurationMs += tableSummary.durationMs;
 
 		(void) IntervalToString(tableSummary.durationMs,
@@ -1077,6 +1269,13 @@ prepare_summary_table(Summary *summary, CopyDataSpec *specs)
 
 		SourceIndexArray indexArray = { 0 };
 
+		/* make sure to always initialize this memory area */
+		entry->indexArray.count = 0;
+		entry->indexArray.array = NULL;
+
+		entry->constraintArray.count = 0;
+		entry->constraintArray.array = NULL;
+
 		/*
 		 * When the table COPY processing was split into several processes,
 		 * ensure we only read the index list once: only one of those COPY
@@ -1087,7 +1286,24 @@ prepare_summary_table(Summary *summary, CopyDataSpec *specs)
 			if (!read_table_index_file(&indexArray,
 									   tableSpecs->tablePaths.idxListFile))
 			{
-				/* errors have already been logged */
+				log_error("Failed to read table index file \"%s\"",
+						  tableSpecs->tablePaths.idxListFile);
+				return false;
+			}
+
+			/* prepare for as many constraints as indexes */
+			entry->indexArray.array =
+				(SummaryIndexEntry *) calloc(indexArray.count,
+											 sizeof(SummaryIndexEntry));
+
+			entry->constraintArray.array =
+				(SummaryIndexEntry *) calloc(indexArray.count,
+											 sizeof(SummaryIndexEntry));
+
+			if (entry->indexArray.array == NULL ||
+				entry->constraintArray.array == NULL)
+			{
+				log_error(ALLOCATION_FAILED_ERROR);
 				return false;
 			}
 
@@ -1108,33 +1324,45 @@ prepare_summary_table(Summary *summary, CopyDataSpec *specs)
 				/* when a table has no indexes, the file doesn't exists */
 				if (file_exists(indexPaths.doneFile))
 				{
-					CopyIndexSummary indexSummary = { .index = index };
+					SummaryIndexEntry *indexEntry =
+						&(entry->indexArray.array[(entry->indexArray.count)++]);
 
-					if (!read_index_summary(&indexSummary, indexPaths.doneFile))
+					if (!summary_read_index_donefile(index,
+													 indexPaths.doneFile,
+													 false, /* constraint */
+													 indexEntry))
 					{
-						/* errors have already been logged */
+						log_error("Failed to read index done file \"%s\"",
+								  indexPaths.doneFile);
 						return false;
 					}
 
 					/* accumulate total duration of creating all the indexes */
-					timings->indexDurationMs += indexSummary.durationMs;
-					indexingDurationMs += indexSummary.durationMs;
+					timings->indexDurationMs += indexEntry->durationMs;
+					indexingDurationMs += indexEntry->durationMs;
 				}
 
 				if (file_exists(indexPaths.constraintDoneFile))
 				{
-					CopyIndexSummary indexSummary = { .index = index };
+					SummaryIndexArray *constraintArray =
+						&(entry->constraintArray);
 
-					if (!read_index_summary(&indexSummary,
-											indexPaths.constraintDoneFile))
+					SummaryIndexEntry *indexEntry =
+						&(constraintArray->array[(constraintArray->count)++]);
+
+					if (!summary_read_index_donefile(index,
+													 indexPaths.constraintDoneFile,
+													 true, /* constraint */
+													 indexEntry))
 					{
-						/* errors have already been logged */
+						log_error("Failed to read index done file \"%s\"",
+								  indexPaths.constraintDoneFile);
 						return false;
 					}
 
 					/* accumulate total duration of creating all the indexes */
-					timings->indexDurationMs += indexSummary.durationMs;
-					indexingDurationMs += indexSummary.durationMs;
+					timings->indexDurationMs += indexEntry->durationMs;
+					indexingDurationMs += indexEntry->durationMs;
 				}
 			}
 		}
@@ -1148,6 +1376,8 @@ prepare_summary_table(Summary *summary, CopyDataSpec *specs)
 		(void) IntervalToString(indexingDurationMs,
 								entry->indexMs,
 								sizeof(entry->indexMs));
+
+		entry->durationIndexMs = indexingDurationMs;
 	}
 
 	/*
@@ -1159,7 +1389,8 @@ prepare_summary_table(Summary *summary, CopyDataSpec *specs)
 
 		if (!read_blobs_summary(&blobsSummary, specs->cfPaths.done.blobs))
 		{
-			/* errors have already been logged */
+			log_error("Failed to read blog summary file \"%s\"",
+					  specs->cfPaths.done.blobs);
 			return false;
 		}
 
@@ -1172,3 +1403,60 @@ prepare_summary_table(Summary *summary, CopyDataSpec *specs)
 
 	return true;
 }
+
+
+/*
+ * summary_read_index_donefile reads a donefile for an index and populates the
+ * information found in the SummaryIndexEntry structure.
+ */
+bool
+summary_read_index_donefile(SourceIndex *index,
+							const char *filename,
+							bool constraint,
+							SummaryIndexEntry *indexEntry)
+{
+	CopyIndexSummary indexSummary = { .index = index };
+
+	if (!read_index_summary(&indexSummary, filename))
+	{
+		/* errors have already been logged */
+		return false;
+	}
+
+	if (constraint)
+	{
+		indexEntry->oid = indexSummary.index->constraintOid;
+
+		IntString oidString = intToString(indexSummary.index->constraintOid);
+		strlcpy(indexEntry->oidStr,
+				oidString.strValue,
+				sizeof(indexEntry->oidStr));
+	}
+	else
+	{
+		indexEntry->oid = indexSummary.index->indexOid;
+
+		IntString oidString = intToString(indexSummary.index->indexOid);
+		strlcpy(indexEntry->oidStr,
+				oidString.strValue,
+				sizeof(indexEntry->oidStr));
+	}
+
+	strlcpy(indexEntry->nspname,
+			indexSummary.index->indexNamespace,
+			sizeof(indexEntry->nspname));
+
+	strlcpy(indexEntry->relname,
+			indexSummary.index->indexRelname,
+			sizeof(indexEntry->relname));
+
+	indexEntry->sql = strdup(indexSummary.command);
+
+	indexEntry->durationMs = indexSummary.durationMs;
+
+	(void) IntervalToString(indexSummary.durationMs,
+							indexEntry->indexMs,
+							sizeof(indexEntry->indexMs));
+
+	return true;
+}
