@@ -10,12 +10,8 @@ set -e
 #  - PGCOPYDB_TABLE_JOBS
 #  - PGCOPYDB_INDEX_JOBS
 
-#
-# pgcopydb list tables include a retry loop, so we use that as a proxy to
-# depend on the source/target Postgres images to be ready
-#
-pgcopydb list tables --source ${PGCOPYDB_SOURCE_PGURI}
-pgcopydb list tables --source ${PGCOPYDB_TARGET_PGURI}
+# make sure source and target databases are ready
+pgcopydb ping
 
 psql -d ${PGCOPYDB_SOURCE_PGURI} -1 -f /usr/src/pgcopydb/import.sql
 
