@@ -10,16 +10,22 @@ set -e
 #  - PGCOPYDB_TABLE_JOBS
 #  - PGCOPYDB_INDEX_JOBS
 
-#
-# pgcopydb list tables include a retry loop, so we use that as a proxy to
-# depend on the source/target Postgres images to be ready
-#
-pgcopydb list tables --source "${PGCOPYDB_SOURCE_PGURI}"
+# make sure source and target databases are ready
+pgcopydb ping
 
-psql -q -d "${PGCOPYDB_SOURCE_PGURI}" -1 -f ./setup/setup.sql
+psql -a -d "${PGCOPYDB_SOURCE_PGURI}" -1 -f ./setup/setup.sql
+
+# create the target needed collation manually for the test
+psql -a -d "${PGCOPYDB_TARGET_PGURI}" -1 <<EOF
+create collation if not exists mycol
+ (
+   locale = 'fr-FR-x-icu',
+   provider = 'icu'
+ );
+EOF
 
 # pgcopydb fork uses the environment variables
-pgcopydb fork
+pgcopydb fork --skip-collations --fail-fast --notice
 
 # now compare the output of running the SQL command with what's expected
 # as we're not root when running tests, can't write in /usr/src
@@ -38,3 +44,16 @@ do
     test -f $e || cat $r
     diff $e $r || exit 1
 done
+
+
+for f in ./script/*.sh
+do
+    t=`basename $f .sh`
+    r=/tmp/results/${t}.out
+    e=./expected/${t}.out
+    bash $f > $r
+    test -f $e || cat $r
+    # exclude logs, whitespaces and blank lines
+    DIFFOPTS='-B -w -I INFO -I WARN'
+    diff ${DIFFOPTS} $e $r || exit 1
+done
